"""
Counterfactual Examples Module for Logos Theory

This module provides counterfactual-specific examples for the logos semantic framework,
including both countermodels showing invalidity and theorems showing validity.

Example Categories:
------------------
1. Counterfactual Logic Countermodels (CF_CM_*):
   - Tests for invalid counterfactual arguments
   - Examples showing where counterfactual principles fail

2. Counterfactual Logic Theorems (CF_TH_*):
   - Tests for valid counterfactual arguments
   - Counterfactual conditional principles

Usage:
------
This module can be run directly with model-checker or dev_cli.py:

```bash
model-checker path/to/this/counterfactual.py
# or in development:
./dev_cli.py path/to/this/counterfactual.py
```

To use a specific collection of examples, modify the example_range dictionary below.
"""

# Standard imports
import sys
import os

# Add parent directories to path for proper imports
current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(current_dir)
parent_parent_dir = os.path.dirname(parent_dir)
if parent_dir not in sys.path:
    sys.path.insert(0, parent_dir)
if parent_parent_dir not in sys.path:
    sys.path.insert(0, parent_parent_dir)

# Import semantic classes
from ...semantic import (
    LogosSemantics,
    LogosProposition,
    LogosModelStructure,
)

# Import operators
from ...operators import LogosOperatorRegistry

# CF_CM_1: COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM_1_premises = ['\\neg A', '(A \\boxright C)']
CF_CM_1_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM_1_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 3,
    'expectation': True,
}
CF_CM_1_example = [
    CF_CM_1_premises,
    CF_CM_1_conclusions,
    CF_CM_1_settings,
]

# CF_CM_2: MIGHT COUNTERFACTUAL ANTECEDENT STRENGTHENING
CF_CM_2_premises = ['\\neg A', '(A \\diamondright C)']
CF_CM_2_conclusions = ['((A \\wedge B) \\diamondright C)']
CF_CM_2_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_2_example = [
    CF_CM_2_premises,
    CF_CM_2_conclusions,
    CF_CM_2_settings,
]

# CF_CM_3: COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH POSSIBILITY
CF_CM_3_premises = ['\\neg A', '(A \\boxright C)', '\\Diamond (A \\wedge B)']
CF_CM_3_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM_3_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_3_example = [
    CF_CM_3_premises,
    CF_CM_3_conclusions,
    CF_CM_3_settings,
]

# CF_CM_4: COUNTERFACTUAL ANTECEDENT STRENGTHENING WITH NEGATION
CF_CM_4_premises = ['\\neg A','(A \\boxright C)']
CF_CM_4_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM_4_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_4_example = [
    CF_CM_4_premises,
    CF_CM_4_conclusions,
    CF_CM_4_settings,
]

# CF_CM_5: COUNTERFACTUAL DOUBLE ANTECEDENT STRENGTHENING
CF_CM_5_premises = ['(A \\boxright C)','(B \\boxright C)']
CF_CM_5_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM_5_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_5_example = [
    CF_CM_5_premises,
    CF_CM_5_conclusions,
    CF_CM_5_settings,
]

# CF_CM_6: WEAKENED MONOTONICITY
CF_CM_6_premises = ['\\neg A', '(A \\boxright B)','(A \\boxright C)']
CF_CM_6_conclusions = ['((A \\wedge B) \\boxright C)']
CF_CM_6_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_6_example = [
    CF_CM_6_premises,
    CF_CM_6_conclusions,
    CF_CM_6_settings,
]

# CF_CM_7: COUNTERFACTUAL CONTRAPOSITION
CF_CM_7_premises = ['(A \\boxright B)']
CF_CM_7_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM_7_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_7_example = [
    CF_CM_7_premises,
    CF_CM_7_conclusions,
    CF_CM_7_settings,
]

# CF_CM_8: COUNTERFACTUAL CONTRAPOSITION WITH NEGATION
CF_CM_8_premises = ['\\neg B','(A \\boxright B)']
CF_CM_8_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM_8_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_8_example = [
    CF_CM_8_premises,
    CF_CM_8_conclusions,
    CF_CM_8_settings,
]

# CF_CM_9: COUNTERFACTUAL CONTRAPOSITION WITH TWO NEGATIONS
CF_CM_9_premises = ['\\neg A','\\neg B','(A \\boxright B)']
CF_CM_9_conclusions = ['(\\neg B \\boxright \\neg A)']
CF_CM_9_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_9_example = [
    CF_CM_9_premises,
    CF_CM_9_conclusions,
    CF_CM_9_settings,
]

# CF_CM_10: TRANSITIVITY
CF_CM_10_premises = ['(A \\boxright B)','(B \\boxright C)']
CF_CM_10_conclusions = ['(A \\boxright C)']
CF_CM_10_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_10_example = [
    CF_CM_10_premises,
    CF_CM_10_conclusions,
    CF_CM_10_settings,
]

# CF_CM_11: COUNTERFACTUAL TRANSITIVITY WITH NEGATION
CF_CM_11_premises = ['\\neg A','(A \\boxright B)','(B \\boxright C)']
CF_CM_11_conclusions = ['(A \\boxright C)']
CF_CM_11_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_11_example = [
    CF_CM_11_premises,
    CF_CM_11_conclusions,
    CF_CM_11_settings,
]

# CF_CM_12: COUNTERFACTUAL TRANSITIVITY WITH TWO NEGATIONS
CF_CM_12_premises = ['\\neg A','\\neg B','(A \\boxright B)','(B \\boxright C)']
CF_CM_12_conclusions = ['(A \\boxright C)']
CF_CM_12_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_12_example = [
    CF_CM_12_premises,
    CF_CM_12_conclusions,
    CF_CM_12_settings,
]

# CF_CM_13: SOBEL SEQUENCE
CF_CM_13_premises = [
    '(A \\boxright X)',
    '\\neg ((A \\wedge B) \\boxright X)',
    '(((A \\wedge B) \\wedge C) \\boxright X)',
    '\\neg ((((A \\wedge B) \\wedge C) \\wedge D) \\boxright X)',
    '(((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\boxright X)',
    '\\neg ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\boxright X)',
    '(((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G) \\boxright X)',
]
CF_CM_13_conclusions = []
CF_CM_13_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_13_example = [
    CF_CM_13_premises,
    CF_CM_13_conclusions,
    CF_CM_13_settings,
]

# CF_CM_14: SOBEL SEQUENCE WITH POSSIBILITY
CF_CM_14_premises = [
    '\\Diamond A',
    '(A \\boxright X)',
    '\\Diamond (A \\wedge B)',
    '\\neg ((A \\wedge B) \\boxright X)',
    '\\Diamond ((A \\wedge B) \\wedge C)',
    '(((A \\wedge B) \\wedge C) \\boxright X)',
    '\\Diamond (((A \\wedge B) \\wedge C) \\wedge D)',
    '\\neg ((((A \\wedge B) \\wedge C) \\wedge D) \\boxright X)',
    '\\Diamond ((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E)',
    '(((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\boxright X)',
    '\\Diamond (((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F)',
    '\\neg ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\boxright X)',
    '\\Diamond ((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G)',
    '(((((((A \\wedge B) \\wedge C) \\wedge D) \\wedge E) \\wedge F) \\wedge G) \\boxright X)',
]
CF_CM_14_conclusions = []
CF_CM_14_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_14_example = [
    CF_CM_14_premises,
    CF_CM_14_conclusions,
    CF_CM_14_settings,
]

# CF_CM_15: COUNTERFACTUAL EXCLUDED MIDDLE
CF_CM_15_premises = ['\\neg A']
CF_CM_15_conclusions = ['(A \\boxright B)','(A \\boxright \\neg B)']
CF_CM_15_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_15_example = [
    CF_CM_15_premises,
    CF_CM_15_conclusions,
    CF_CM_15_settings,
]

# CF_CM_16: SIMPLIFICATION OF DISJUNCTIVE CONSEQUENT
CF_CM_16_premises = ['\\neg A','(A \\boxright (B \\vee C))']
CF_CM_16_conclusions = ['(A \\boxright B)','(A \\boxright C)']
CF_CM_16_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_16_example = [
    CF_CM_16_premises,
    CF_CM_16_conclusions,
    CF_CM_16_settings,
]

# CF_CM_17: INTRODUCTION OF DISJUNCTIVE ANTECEDENT
CF_CM_17_premises = ['(A \\boxright C)','(B \\boxright C)']
CF_CM_17_conclusions = ['((A \\vee B) \\boxright C)']
CF_CM_17_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_17_example = [
    CF_CM_17_premises,
    CF_CM_17_conclusions,
    CF_CM_17_settings,
]

# CF_CM_18: MUST FACTIVITY
CF_CM_18_premises = ['A','B']
CF_CM_18_conclusions = ['(A \\boxright B)']
CF_CM_18_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_18_example = [
    CF_CM_18_premises,
    CF_CM_18_conclusions,
    CF_CM_18_settings,
]

# CF_CM_19: COUNTERFACTUAL EXPORTATION
CF_CM_19_premises = ['((A \\wedge B) \\boxright C)']
CF_CM_19_conclusions = ['(A \\boxright (B \\boxright C))']
CF_CM_19_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 3,
    'iterate': 1,
    'expectation': True,
}
CF_CM_19_example = [
    CF_CM_19_premises,
    CF_CM_19_conclusions,
    CF_CM_19_settings,
]

# CF_CM_20: COUNTERFACTUAL EXPORTATION WITH POSSIBILITY
CF_CM_20_premises = ['((A \\wedge B) \\boxright C)','\\Diamond (A \\wedge B)']
CF_CM_20_conclusions = ['(A \\boxright (B \\boxright C))']
CF_CM_20_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 3,
    'iterate': 1,
    'expectation': True,
}
CF_CM_20_example = [
    CF_CM_20_premises,
    CF_CM_20_conclusions,
    CF_CM_20_settings,
]

# CF_CM_21: COUNTERFACTUAL NEGATION DISTRIBUTION
CF_CM_21_premises = ['\\neg A','\\neg (A \\boxright B)']
CF_CM_21_conclusions = ['(A \\boxright \\neg B)']
CF_CM_21_settings = {
    'N': 4,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 2,
    'iterate': 1,
    'expectation': True,
}
CF_CM_21_example = [
    CF_CM_21_premises,
    CF_CM_21_conclusions,
    CF_CM_21_settings,
]

# CF_CM_24: CONTRADICTION TO IMPOSSIBILITY
CF_CM_24_premises = ['(A \\boxright B)']
CF_CM_24_conclusions = ['\\Box (A \\boxright B)']
CF_CM_24_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_24_example = [
    CF_CM_24_premises,
    CF_CM_24_conclusions,
    CF_CM_24_settings,
]

# CF_CM_25: CONTRADICTION TO IMPOSSIBILITY
CF_CM_25_premises = ['A', '\\Diamond B', '\\neg \\Diamond (A \\wedge B)']
CF_CM_25_conclusions = ['(B \\boxright C)']
CF_CM_25_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': True,
}
CF_CM_25_example = [
    CF_CM_25_premises,
    CF_CM_25_conclusions,
    CF_CM_25_settings,
]



##################
### THEOREMS  ###
##################

# CF_TH_1: COUNTERFACTUAL IDENTITY
CF_TH_1_premises = []
CF_TH_1_conclusions = ['(A \\boxright A)']
CF_TH_1_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_1_example = [
    CF_TH_1_premises,
    CF_TH_1_conclusions,
    CF_TH_1_settings,
]

# CF_TH_2: COUNTERFACTUAL MODUS PONENS
CF_TH_2_premises = ['A','(A \\boxright B)']
CF_TH_2_conclusions = ['B']
CF_TH_2_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_2_example = [
    CF_TH_2_premises,
    CF_TH_2_conclusions,
    CF_TH_2_settings,
]

# CF_TH_3: WEAKENED TRANSITIVITY
CF_TH_3_premises = ['(A \\boxright B)','((A \\wedge B) \\boxright C)']
CF_TH_3_conclusions = ['(A \\boxright C)']
CF_TH_3_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_3_example = [
    CF_TH_3_premises,
    CF_TH_3_conclusions,
    CF_TH_3_settings,
]

# CF_TH_4: ANTECEDENT DISJUNCTION TO CONJUNCTION
CF_TH_4_premises = ['((A \\vee B) \\boxright C)']
CF_TH_4_conclusions = ['((A \\wedge B) \\boxright C)']
CF_TH_4_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_4_example = [
    CF_TH_4_premises,
    CF_TH_4_conclusions,
    CF_TH_4_settings,
]

# CF_TH_5: SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT
CF_TH_5_premises = ['((A \\vee B) \\boxright C)']
CF_TH_5_conclusions = ['(A \\boxright C)']
CF_TH_5_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_5_example = [
    CF_TH_5_premises,
    CF_TH_5_conclusions,
    CF_TH_5_settings,
]

# CF_TH_6: DOUBLE SIMPLIFICATION OF DISJUNCTIVE ANTECEDENT
CF_TH_6_premises = ['((A \\vee B) \\boxright C)']
CF_TH_6_conclusions = ['((A \\boxright C) \\wedge (B \\boxright C))']
CF_TH_6_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_6_example = [
    CF_TH_6_premises,
    CF_TH_6_conclusions,
    CF_TH_6_settings,
]

# CF_TH_7: COUNTERFACTUAL DISJUNCTION INTRODUCTION
CF_TH_7_premises = [
    '(A \\boxright C)',
    '(B \\boxright C)',
    '((A \\wedge B) \\boxright C)',
]
CF_TH_7_conclusions = ['((A \\vee B) \\boxright C)']
CF_TH_7_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_7_example = [
    CF_TH_7_premises,
    CF_TH_7_conclusions,
    CF_TH_7_settings,
]

# CF_TH_8: COUNTERFACTUAL CONSEQUENT WEAKENING
CF_TH_8_premises = ['(A \\boxright (B \\wedge C))']
CF_TH_8_conclusions = ['(A \\boxright B)']
CF_TH_8_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_8_example = [
    CF_TH_8_premises,
    CF_TH_8_conclusions,
    CF_TH_8_settings,
]

# CF_TH_9: COUNTERFACTUAL CONJUNCTION INTRODUCTION
CF_TH_9_premises = ['(A \\boxright B)','(A \\boxright C)']
CF_TH_9_conclusions = ['(A \\boxright (B \\wedge C))']
CF_TH_9_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_9_example = [
    CF_TH_9_premises,
    CF_TH_9_conclusions,
    CF_TH_9_settings,
]

# CF_TH_10: MIGHT FACTIVITY
CF_TH_10_premises = ['A','B']
CF_TH_10_conclusions = ['(A \\diamondright B)']
CF_TH_10_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_10_example = [
    CF_TH_10_premises,
    CF_TH_10_conclusions,
    CF_TH_10_settings,
]
# CF_TH_11: DEFINITION OF NEC
CF_TH_11_premises = ['\\Box A']
CF_TH_11_conclusions = ['(\\top \\boxright A)']
CF_TH_11_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': True,
    'non_null': True,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_11_example = [
    CF_TH_11_premises,
    CF_TH_11_conclusions,
    CF_TH_11_settings,
]

# CF_TH_12: CONTRADICTION TO IMPOSSIBILITY
CF_TH_12_premises = ['(A \\boxright \\bot)']
CF_TH_12_conclusions = ['(\\top \\boxright \\neg A)']
CF_TH_12_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 10,
    'iterate': 1,
    'expectation': False,
}
CF_TH_12_example = [
    CF_TH_12_premises,
    CF_TH_12_conclusions,
    CF_TH_12_settings,
]

# CF_TH_13: REVERSE DEFINITION OF NEC
CF_TH_13_premises = ['(\\top \\boxright A)']
CF_TH_13_conclusions = ['\\Box A']
CF_TH_13_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': True,
    'non_null': True,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
CF_TH_13_example = [
    CF_TH_13_premises,
    CF_TH_13_conclusions,
    CF_TH_13_settings,
]

# CF_CM_14: CONTRADICTION TO IMPOSSIBILITY
CF_TH_14_premises = ['(A \\boxright \\bot)']
CF_TH_14_conclusions = ['(\\top \\boxright \\neg A)']
CF_TH_14_settings = {
    'N': 4,
    'contingent': False,
    'disjoint': False,
    'non_empty': False,
    'non_null': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
CF_TH_14_example = [
    CF_TH_14_premises,
    CF_TH_14_conclusions,
    CF_TH_14_settings,
]


# Create collections for different counterfactual example types
counterfactual_cm_examples = {
    "CF_CM_1": CF_CM_1_example,
    "CF_CM_2": CF_CM_2_example,
    "CF_CM_3": CF_CM_3_example,
    "CF_CM_4": CF_CM_4_example,
    "CF_CM_5": CF_CM_5_example,
    "CF_CM_6": CF_CM_6_example,
    "CF_CM_7": CF_CM_7_example,
    "CF_CM_8": CF_CM_8_example,
    "CF_CM_9": CF_CM_9_example,
    "CF_CM_10": CF_CM_10_example,
    "CF_CM_11": CF_CM_11_example,
    "CF_CM_12": CF_CM_12_example,
    "CF_CM_13": CF_CM_13_example,
    "CF_CM_14": CF_CM_14_example,
    "CF_CM_15": CF_CM_15_example,
    "CF_CM_16": CF_CM_16_example,
    "CF_CM_17": CF_CM_17_example,
    "CF_CM_18": CF_CM_18_example,
    "CF_CM_19": CF_CM_19_example,
    "CF_CM_20": CF_CM_20_example,
    "CF_CM_21": CF_CM_21_example,
    "CF_CM_24": CF_CM_24_example,
    "CF_CM_25": CF_CM_25_example,
}

counterfactual_th_examples = {
    "CF_TH_1": CF_TH_1_example,
    "CF_TH_2": CF_TH_2_example,
    "CF_TH_3": CF_TH_3_example,
    "CF_TH_4": CF_TH_4_example,
    "CF_TH_5": CF_TH_5_example,
    "CF_TH_6": CF_TH_6_example,
    "CF_TH_7": CF_TH_7_example,
    "CF_TH_8": CF_TH_8_example,
    "CF_TH_9": CF_TH_9_example,
    "CF_TH_10": CF_TH_10_example,
    "CF_TH_11": CF_TH_11_example,
    "CF_TH_12": CF_TH_12_example,
    "CF_TH_13": CF_TH_13_example,
    "CF_TH_14": CF_TH_14_example,
}

# Combined collection of all counterfactual examples
unit_tests = {**counterfactual_cm_examples, **counterfactual_th_examples}

# Default settings
general_settings = {
    "print_constraints": False,
    "print_impossible": True,
    "print_z3": False,
    "save_output": False,
    "maximize": False,
}

# Create operator registry for counterfactual theory (includes modal for Box and Diamond operators)
counterfactual_registry = LogosOperatorRegistry()
counterfactual_registry.load_subtheories(['extensional', 'modal', 'counterfactual'])

counterfactual_theory = {
    "semantics": LogosSemantics,
    "proposition": LogosProposition,
    "model": LogosModelStructure,
    "operators": counterfactual_registry.get_operators(),
}

# Specify which theories to use
semantic_theories = {
    "Brast-McKie": counterfactual_theory,
}

# # Specify which examples to run by default when running this module directly
# # All examples included by default
# example_range = unit_tests

# Or set specific examples
example_range = {
    "CF_CM_1": CF_CM_1_example,
    "CF_TH_5": CF_TH_5_example,
}


def get_examples():
    """
    Get all counterfactual examples.
    
    Returns:
        dict: Dictionary containing all counterfactual examples
    """
    return {
        'countermodels': counterfactual_cm_examples,
        'theorems': counterfactual_th_examples,
        'all': unit_tests
    }

# Make this module runnable from the command line
if __name__ == '__main__':
    import subprocess
    file_name = os.path.basename(__file__)
    subprocess.run(["model-checker", file_name], check=True, cwd=parent_parent_dir)



"""Tests for the validation utilities."""

import unittest
from unittest.mock import Mock, MagicMock, patch

from model_checker.models.semantic import SemanticDefaults
from model_checker.models.proposition import PropositionDefaults
from model_checker.models.structure import ModelDefaults
from model_checker.syntactic import OperatorCollection
from model_checker.builder.validation import (
    validate_semantic_theory,
    validate_example_case
)

# Mock classes for testing
class MockSemantics(SemanticDefaults):
    def __init__(self, *args, **kwargs):
        pass

class MockProposition(PropositionDefaults):
    def __init__(self, *args, **kwargs):
        pass

class MockModel(ModelDefaults):
    def __init__(self, *args, **kwargs):
        pass

class TestValidation(unittest.TestCase):
    """Test the validation utilities."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.mock_operators = Mock(spec=OperatorCollection)
        
        # Create valid semantic theory
        self.valid_theory = {
            "semantics": MockSemantics,
            "proposition": MockProposition,
            "operators": self.mock_operators,
            "model": MockModel
        }
        
        # Create valid example case
        self.valid_example = (
            ["premise1", "premise2"],
            ["conclusion"],
            {"N": 3, "max_time": 30}
        )
    
    def test_validate_semantic_theory_valid(self):
        """Test validating a valid semantic theory."""
        result = validate_semantic_theory(self.valid_theory)
        self.assertEqual(result[0], MockSemantics)
        self.assertEqual(result[1], MockProposition)
        self.assertEqual(result[2], self.mock_operators)
        self.assertEqual(result[3], MockModel)
    
    def test_validate_semantic_theory_not_dict(self):
        """Test validating a non-dictionary semantic theory."""
        with self.assertRaises(TypeError) as context:
            validate_semantic_theory(["not", "a", "dict"])
        self.assertTrue("must be a dictionary" in str(context.exception))
    
    def test_validate_semantic_theory_missing_component(self):
        """Test validating a semantic theory with missing component."""
        for key in self.valid_theory.keys():
            invalid_theory = self.valid_theory.copy()
            del invalid_theory[key]
            
            with self.assertRaises(TypeError) as context:
                validate_semantic_theory(invalid_theory)
            self.assertTrue(f"missing required component: {key}" in str(context.exception))
    
    def test_validate_semantic_theory_invalid_types(self):
        """Test validating a semantic theory with invalid component types."""
        # Test invalid semantics
        invalid_theory = self.valid_theory.copy()
        invalid_theory["semantics"] = object  # Not a SemanticDefaults subclass
        
        with self.assertRaises(TypeError) as context:
            validate_semantic_theory(invalid_theory)
        self.assertTrue("'semantics' must be a subclass of SemanticDefaults" in str(context.exception))
        
        # Test invalid proposition
        invalid_theory = self.valid_theory.copy()
        invalid_theory["proposition"] = object  # Not a PropositionDefaults subclass
        
        with self.assertRaises(TypeError) as context:
            validate_semantic_theory(invalid_theory)
        self.assertTrue("'proposition' must be a subclass of PropositionDefaults" in str(context.exception))
        
        # Test invalid operators
        invalid_theory = self.valid_theory.copy()
        invalid_theory["operators"] = {}  # Not an OperatorCollection instance
        
        with self.assertRaises(TypeError) as context:
            validate_semantic_theory(invalid_theory)
        self.assertTrue("'operators' must be an instance of OperatorCollection" in str(context.exception))
        
        # Test invalid model
        invalid_theory = self.valid_theory.copy()
        invalid_theory["model"] = object  # Not a ModelDefaults subclass
        
        with self.assertRaises(TypeError) as context:
            validate_semantic_theory(invalid_theory)
        self.assertTrue("'model' must be a subclass of ModelDefaults" in str(context.exception))
    
    def test_validate_example_case_valid(self):
        """Test validating a valid example case."""
        result = validate_example_case(self.valid_example)
        self.assertEqual(result[0], ["premise1", "premise2"])
        self.assertEqual(result[1], ["conclusion"])
        self.assertEqual(result[2], {"N": 3, "max_time": 30})
    
    def test_validate_example_case_not_list(self):
        """Test validating a non-list example case."""
        with self.assertRaises(TypeError) as context:
            validate_example_case("not a list")
        self.assertTrue("must be a list/tuple" in str(context.exception))
    
    def test_validate_example_case_wrong_length(self):
        """Test validating an example case with wrong number of elements."""
        with self.assertRaises(TypeError) as context:
            validate_example_case(["too", "few"])
        self.assertTrue("must be a list/tuple containing exactly 3 elements" in str(context.exception))
    
    def test_validate_example_case_invalid_premises(self):
        """Test validating an example case with invalid premises."""
        invalid_example = (123, ["conclusion"], {})  # First element not a list of strings
        
        with self.assertRaises(TypeError) as context:
            validate_example_case(invalid_example)
        self.assertTrue("premises must be a list of strings" in str(context.exception))
        
        invalid_example = ([1, 2, 3], ["conclusion"], {})  # List but not of strings
        
        with self.assertRaises(TypeError) as context:
            validate_example_case(invalid_example)
        self.assertTrue("premises must be a list of strings" in str(context.exception))
    
    def test_validate_example_case_invalid_conclusions(self):
        """Test validating an example case with invalid conclusions."""
        invalid_example = (["premise"], 123, {})  # Second element not a list of strings
        
        with self.assertRaises(TypeError) as context:
            validate_example_case(invalid_example)
        self.assertTrue("conclusions must be a list of strings" in str(context.exception))
        
        invalid_example = (["premise"], [1, 2, 3], {})  # List but not of strings
        
        with self.assertRaises(TypeError) as context:
            validate_example_case(invalid_example)
        self.assertTrue("conclusions must be a list of strings" in str(context.exception))
    
    def test_validate_example_case_invalid_settings(self):
        """Test validating an example case with invalid settings."""
        invalid_example = (["premise"], ["conclusion"], "not a dict")  # Third element not a dictionary
        
        with self.assertRaises(TypeError) as context:
            validate_example_case(invalid_example)
        self.assertTrue("settings must be a dictionary" in str(context.exception))

if __name__ == "__main__":
    unittest.main()


"""Tests for the bimodal iteration system.

This module tests the BimodalModelIterator implementation to ensure it can
correctly find multiple distinct models for bimodal logic examples.
"""

import pytest
from model_checker.theory_lib.bimodal import (
    BimodalSemantics,
    BimodalProposition,
    BimodalStructure,
    bimodal_operators,
    BimodalModelIterator,
    iterate_example,
)
from model_checker.builder.example import BuildExample
from model_checker.models.constraints import ModelConstraints
from model_checker.syntactic import Syntax


@pytest.mark.skip(reason="Test requires a properly initialized BuildModule which is difficult to mock")
def test_basic_iteration():
    """Test basic iteration for a simple bimodal formula."""
    # Create a simple formula that has multiple models
    premises = ['(A \\vee B)']
    conclusions = []
    settings = {
        'N': 1,
        'M': 2,
        'contingent': False,
        'disjoint': False,
        'max_time': 5,
        'expectation': True,
        'iterate': 3,  # Find up to 3 distinct models
    }

    # Original test implementation would go here
    # We would create a BuildExample, check the initial model is valid,
    # create a BimodalModelIterator, and verify we can find multiple models
    pass


@pytest.mark.skip(reason="Test requires a properly initialized BuildModule which is difficult to mock")
def test_iterate_example_function():
    """Test the iterate_example function."""
    # Create a simple formula that has multiple models
    premises = ['(A \\vee B)']
    conclusions = []
    settings = {
        'N': 1,
        'M': 2,
        'contingent': False,
        'disjoint': False,
        'max_time': 5,
        'expectation': True,
    }

    # Original test implementation would go here
    # We would create a BuildExample, check the initial model is valid,
    # use iterate_example to find multiple models, and verify we found them
    pass


@pytest.mark.skip(reason="Test requires a properly initialized BuildModule which is difficult to mock")
def test_bimodal_specific_differences():
    """Test that bimodal-specific differences are captured."""
    # Create a simple formula that has multiple models with different world histories
    premises = ['(A \\vee B)']
    conclusions = []
    settings = {
        'N': 1,
        'M': 2,
        'contingent': False,
        'disjoint': False,
        'max_time': 5,
        'expectation': True,
        'iterate': 3,  # Find up to 3 distinct models
    }

    # Original test implementation would go here
    # We would create a BuildExample, create a BimodalModelIterator,
    # find multiple models, and check for bimodal-specific differences
    pass


"""Tests for Z3 solver state isolation between examples.

This test suite verifies that solver state isolation works correctly,
ensuring examples produce consistent results regardless of execution order.
"""

import unittest
import sys
import os
import z3
from model_checker.builder import BuildModule
from model_checker.utils import Z3ContextManager

class TestZ3ContextIsolation(unittest.TestCase):
    """Test cases for Z3 context isolation."""

    def setUp(self):
        """Set up test environment."""
        # Make sure we start with a fresh context
        Z3ContextManager.reset_context()
        
        # Minimal settings for a build module
        self.test_settings = {
            "file_path": "none.py",  # Not actually used in these tests
        }
        
        # Create a dummy z3 formula to check solver behavior
        self.x = z3.Int('x')
        self.y = z3.Int('y')
        
    def test_context_reset(self):
        """Test that the Z3 context reset mechanism works correctly."""
        # Instead of testing internal Z3 implementation details, 
        # test the behavior from a black-box perspective
        
        # Create a solver with a constraint
        solver1 = z3.Solver()
        x = z3.Int('x')
        solver1.add(x > 5)
        
        # Reset context
        Z3ContextManager.reset_context()
        
        # Create a new solver that shouldn't see solver1's constraints
        solver2 = z3.Solver()
        solver2.add(x < 0)
        
        # If context was properly reset, solver2 should be satisfiable
        # despite having a constraint that conflicts with solver1
        result = solver2.check()
        self.assertEqual(result, z3.sat)
        
    def test_solver_state_isolation(self):
        """Test that solver state doesn't leak between invocations."""
        # First solver with simple constraints
        Z3ContextManager.reset_context()
        solver1 = z3.Solver()
        solver1.add(self.x > 0)
        solver1.add(self.x < 10)
        result1 = solver1.check()
        self.assertEqual(result1, z3.sat)
        
        # Get model and assert 0 < x < 10
        model1 = solver1.model()
        x_val = model1.eval(self.x).as_long()
        self.assertTrue(0 < x_val < 10)
        
        # Create a second solver with completely different constraints
        Z3ContextManager.reset_context()
        solver2 = z3.Solver()
        solver2.add(self.y < 0)  # Different variable
        result2 = solver2.check()
        self.assertEqual(result2, z3.sat)
        
        # Second solver should not know about constraints from first solver
        model2 = solver2.model()
        if self.x in model2:  # Sometimes it doesn't even include x
            x_val2 = model2.eval(self.x).as_long()
            self.assertNotEqual(x_val, x_val2)

    def test_conflicting_constraints(self):
        """Test that conflicting constraints in separate solvers don't interfere."""
        # First solver with x > 0
        Z3ContextManager.reset_context()
        solver1 = z3.Solver()
        solver1.add(self.x > 0)
        result1 = solver1.check()
        self.assertEqual(result1, z3.sat)
        
        # Second solver with x < 0 (conflicting with first)
        Z3ContextManager.reset_context()
        solver2 = z3.Solver()
        solver2.add(self.x < 0)
        result2 = solver2.check()
        
        # Despite conflict with first solver's constraint, second should be SAT
        self.assertEqual(result2, z3.sat)

    def test_solver_state_leakage_without_reset(self):
        """Test that solver state leaks without proper resets."""
        # This test demonstrates what happens WITHOUT resetting the context
        
        # First solver with learned information about y = 2*x
        solver1 = z3.Solver()
        solver1.add(self.y == 2 * self.x)
        solver1.add(self.x == 5)
        result1 = solver1.check()
        self.assertEqual(result1, z3.sat)
        
        # Deliberately NOT resetting context
        
        # Create a second solver with only y == 10 constraint
        # This on its own should be SAT with many possible values for x
        solver2 = z3.Solver()
        solver2.add(self.y == 10)
        result2 = solver2.check()
        self.assertEqual(result2, z3.sat)
        
        # Get the model - because solver1 taught the context that y = 2*x and x = 5,
        # we might see those constraints influencing the results
        model2 = solver2.model()
        
        # Now reset context for clean state
        Z3ContextManager.reset_context()
        
        # Create a third solver with only y == 10 constraint after context reset
        solver3 = z3.Solver()
        solver3.add(self.y == 10)
        result3 = solver3.check()
        self.assertEqual(result3, z3.sat)
        
        # Get model - should be different from second solver due to context reset
        model3 = solver3.model()
        
        # The exact values aren't predictable, but the point is that model2 and model3
        # should differ because solver3 has a fresh context with no knowledge of
        # prior constraints, while solver2 might be influenced by solver1's state
        if self.x in model2 and self.x in model3:
            x_val2 = model2.eval(self.x).as_long() 
            x_val3 = model3.eval(self.x).as_long()
            
            # This assertion might occasionally fail if Z3 randomly picks the same
            # values in both cases, but it's unlikely with large integer ranges
            if x_val2 == 5:  # If we see evidence of state leakage
                self.assertNotEqual(x_val3, 5, 
                                 "Context reset didn't prevent state leakage")
            

if __name__ == '__main__':
    unittest.main()


"""Unit tests for the settings module.
"""

import unittest
from model_checker.settings.settings import SettingsManager

class MockSemantics:
    """Mock semantics for testing."""
    DEFAULT_EXAMPLE_SETTINGS = {
        'N': 3,
        'max_time': 1,
        'expectation': None,
    }
    
    DEFAULT_GENERAL_SETTINGS = {
        'print_z3': False,
        'save_output': False,
    }

class MockFlags:
    """Mock flags for testing."""
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

class TestSettingsManager(unittest.TestCase):
    """Test the SettingsManager class."""
    
    def setUp(self):
        """Set up test cases."""
        self.semantic_theory = {
            'semantics': MockSemantics
        }
        self.settings_manager = SettingsManager(self.semantic_theory)
    
    def test_validate_general_settings(self):
        """Test validating general settings."""
        # Test with None
        settings = self.settings_manager.validate_general_settings(None)
        self.assertEqual(settings, MockSemantics.DEFAULT_GENERAL_SETTINGS)
        
        # Test with valid settings
        user_settings = {'print_z3': True}
        settings = self.settings_manager.validate_general_settings(user_settings)
        self.assertEqual(settings['print_z3'], True)
        
        # Test with unknown settings (should print warning but include valid ones)
        user_settings = {'print_z3': True, 'unknown': 'value'}
        settings = self.settings_manager.validate_general_settings(user_settings)
        self.assertEqual(settings['print_z3'], True)
        self.assertNotIn('unknown', settings)
    
    def test_validate_example_settings(self):
        """Test validating example settings."""
        # Test with None
        settings = self.settings_manager.validate_example_settings(None)
        self.assertEqual(settings, MockSemantics.DEFAULT_EXAMPLE_SETTINGS)
        
        # Test with valid settings
        user_settings = {'N': 5}
        settings = self.settings_manager.validate_example_settings(user_settings)
        self.assertEqual(settings['N'], 5)
        
        # Test with unknown settings
        user_settings = {'N': 5, 'unknown': 'value'}
        settings = self.settings_manager.validate_example_settings(user_settings)
        self.assertEqual(settings['N'], 5)
        self.assertNotIn('unknown', settings)
    
    def test_apply_flag_overrides(self):
        """Test applying flag overrides."""
        settings = {'print_z3': False, 'N': 3}
        flags = MockFlags(print_z3=True, unknown=True)
        
        result = self.settings_manager.apply_flag_overrides(settings, flags)
        self.assertEqual(result['print_z3'], True)
        self.assertEqual(result['N'], 3)
    
    def test_get_complete_settings(self):
        """Test getting complete settings."""
        general_settings = {'print_z3': True}
        example_settings = {'N': 5}
        flags = MockFlags(save_output=True)
        
        result = self.settings_manager.get_complete_settings(
            general_settings, example_settings, flags
        )
        
        # Check all levels of settings are applied in correct order
        self.assertEqual(result['print_z3'], True)  # From general_settings
        self.assertEqual(result['N'], 5)  # From example_settings
        self.assertEqual(result['save_output'], True)  # From flags
        self.assertEqual(result['expectation'], None)  # From DEFAULT_EXAMPLE_SETTINGS

if __name__ == '__main__':
    unittest.main()


"""Tests for imposition data extraction methods."""

import unittest
from unittest.mock import Mock

from model_checker.theory_lib.imposition.semantic import ImpositionModelStructure


class TestImpositionDataExtraction(unittest.TestCase):
    """Test data extraction methods for ImpositionModelStructure."""
    
    def test_extract_states(self):
        """Test extraction of categorized states."""
        # Create a mock structure
        structure = Mock(spec=ImpositionModelStructure)
        
        # Add the method we're testing
        structure.extract_states = ImpositionModelStructure.extract_states.__get__(structure)
        
        # Mock state collections
        structure.z3_world_states = [3, 5]  # World states
        structure.z3_possible_states = [1, 3, 5, 7]  # Possible states (including worlds)
        structure.all_states = [0, 1, 2, 3, 4, 5, 6, 7]  # All states
        
        # Test extraction
        states = structure.extract_states()
        
        # Verify result structure
        self.assertIsInstance(states, dict)
        self.assertIn('worlds', states)
        self.assertIn('possible', states)
        self.assertIn('impossible', states)
        
        # Verify categorization
        self.assertEqual(set(states['worlds']), {'s3', 's5'})
        self.assertEqual(set(states['possible']), {'s1', 's7'})  # Excludes worlds
        self.assertEqual(set(states['impossible']), {'s0', 's2', 's4', 's6'})
        
    def test_extract_evaluation_world(self):
        """Test extraction of main evaluation world."""
        structure = Mock(spec=ImpositionModelStructure)
        structure.extract_evaluation_world = ImpositionModelStructure.extract_evaluation_world.__get__(structure)
        
        # Mock semantics with main_world
        structure.semantics = Mock()
        structure.semantics.main_world = 5
        
        # Test extraction
        world = structure.extract_evaluation_world()
        self.assertEqual(world, 's5')
        
        # Test with None
        structure.semantics.main_world = None
        world = structure.extract_evaluation_world()
        self.assertIsNone(world)
        
    def test_extract_relations(self):
        """Test extraction of imposition relations."""
        structure = Mock(spec=ImpositionModelStructure)
        structure.extract_relations = ImpositionModelStructure.extract_relations.__get__(structure)
        
        # Mock imposition relations: (state, world, outcome)
        structure.z3_imposition_relations = [
            (1, 3, 3),  # s3 ->_s1 s3
            (2, 3, 5),  # s3 ->_s2 s5
            (1, 5, 7),  # s5 ->_s1 s7
            (4, 5, 5),  # s5 ->_s4 s5
        ]
        
        # Test extraction
        relations = structure.extract_relations()
        
        # Verify structure
        self.assertIsInstance(relations, dict)
        self.assertIn('imposition', relations)
        
        # Verify content
        impositions = relations['imposition']
        self.assertEqual(impositions['s3'], {'s1': 's3', 's2': 's5'})
        self.assertEqual(impositions['s5'], {'s1': 's7', 's4': 's5'})
        
    def test_extract_propositions(self):
        """Test extraction of proposition values."""
        structure = Mock(spec=ImpositionModelStructure)
        structure.extract_propositions = ImpositionModelStructure.extract_propositions.__get__(structure)
        
        # Mock world states
        structure.z3_world_states = [3, 5]
        
        # Mock syntax with propositions
        mock_syntax = Mock()
        mock_prop_p = Mock()
        mock_prop_p.evaluate_at = Mock(side_effect=lambda w: w == 3)  # True at world 3
        
        mock_prop_q = Mock()
        mock_prop_q.evaluate_at = Mock(side_effect=lambda w: w == 5)  # True at world 5
        
        mock_syntax.propositions = {'p': mock_prop_p, 'q': mock_prop_q}
        structure.syntax = mock_syntax
        
        # Test extraction
        propositions = structure.extract_propositions()
        
        # Verify structure
        self.assertIsInstance(propositions, dict)
        self.assertIn('p', propositions)
        self.assertIn('q', propositions)
        
        # Verify values
        self.assertEqual(propositions['p'], {'s3': True, 's5': False})
        self.assertEqual(propositions['q'], {'s3': False, 's5': True})
        
    def test_extract_propositions_no_syntax(self):
        """Test proposition extraction when no syntax available."""
        structure = Mock(spec=ImpositionModelStructure)
        structure.extract_propositions = ImpositionModelStructure.extract_propositions.__get__(structure)
        
        # No syntax attribute
        structure.syntax = Mock()
        delattr(structure.syntax, 'propositions')
        
        propositions = structure.extract_propositions()
        
        # Should return empty dict
        self.assertEqual(propositions, {})
        
    def test_extract_relations_no_impositions(self):
        """Test relation extraction when no impositions defined."""
        structure = Mock(spec=ImpositionModelStructure)
        structure.extract_relations = ImpositionModelStructure.extract_relations.__get__(structure)
        
        # No imposition relations
        delattr(structure, 'z3_imposition_relations')
        
        relations = structure.extract_relations()
        
        # Should return empty structure
        self.assertEqual(relations, {})


if __name__ == '__main__':
    unittest.main()


"""Tests for the base model iterator functionality."""

import pytest
import z3
import time
from unittest.mock import Mock, patch
from model_checker.iterate.core import BaseModelIterator
from model_checker.builder.example import BuildExample


class MockModelIterator(BaseModelIterator):
    """Mock implementation for testing base functionality."""
    
    def _calculate_differences(self, new_structure, previous_structure):
        return {"test": "differences"}
    
    def _create_difference_constraint(self, previous_models):
        return z3.BoolVal(True)
    
    def _create_non_isomorphic_constraint(self, z3_model):
        return z3.BoolVal(True)
    
    def _create_stronger_constraint(self, isomorphic_model):
        return z3.BoolVal(True)


class TestBaseModelIterator:
    """Test cases for BaseModelIterator."""
    
    @pytest.mark.skip(reason="Temporarily skip - pending BUILDER_IMPORT.md refactoring")
    def test_abstract_methods_required(self):
        """Test that abstract methods must be implemented."""
        with pytest.raises(TypeError):
            BaseModelIterator(Mock())
    
    @pytest.mark.skip(reason="Temporarily skip - pending BUILDER_IMPORT.md refactoring")
    def test_initialization_validation(self):
        """Test initialization validates inputs."""
        # Test with invalid BuildExample
        with pytest.raises(TypeError):
            MockModelIterator("not a BuildExample")
        
        # Test with BuildExample without model
        mock_example = Mock(spec=BuildExample)
        mock_example.model_structure = None
        with pytest.raises(ValueError, match="no model_structure"):
            MockModelIterator(mock_example)
    
    def test_timeout_handling(self):
        """Test iteration timeout is properly handled."""
        # Create mock example with slow solver
        mock_example = create_mock_example()
        iterator = MockModelIterator(mock_example)
        
        # Set very short timeout
        iterator.settings['iteration_timeout'] = 0.001
        
        # Mock slow solver check
        def slow_check():
            import time
            time.sleep(0.1)
            return z3.unsat
        
        with patch.object(iterator.solver, 'check', side_effect=slow_check):
            # Run iteration
            models = iterator.iterate()
        
        # Should stop due to timeout or no models
        assert len(models) == 1  # Only initial model
        # Check messages - unsat should mean no additional models
        debug_msgs = iterator.debug_messages
        # Print for debugging
        print(f"Debug messages: {debug_msgs}")
        # Either timeout or unsat (no models) message should be present
        assert len(debug_msgs) > 0  # Should have some messages
    
    def test_invalid_model_handling(self):
        """Test handling of models with no possible worlds."""
        mock_example = create_mock_example()
        iterator = MockModelIterator(mock_example)
        
        # Mock solver to return sat but with invalid model
        with patch.object(iterator.constraint_generator.solver, 'check', return_value=z3.sat):
            # Mock constraint_generator.get_model() to return a valid Z3 model
            mock_z3_model = Mock()
            with patch.object(iterator.constraint_generator, 'get_model', return_value=mock_z3_model):
                with patch.object(iterator.model_builder, 'build_new_model_structure') as mock_build:
                    # Return structure with no worlds
                    mock_structure = Mock()
                    mock_structure.z3_world_states = []
                    mock_structure.model_differences = {"test": "differences"}
                    mock_build.return_value = mock_structure
                    
                    # Run iteration requesting 3 models
                    iterator.max_iterations = 3
                    models = iterator.iterate()
                    
                    # Should only have initial model
                    assert len(models) == 1
                    # Check for invalid model message
                    debug_msgs = iterator.debug_messages
                    assert any("invalid" in msg.lower() or "no world" in msg.lower()
                              for msg in debug_msgs)
    
    def test_consecutive_invalid_limit(self):
        """Test that consecutive invalid models trigger stop."""
        mock_example = create_mock_example()
        iterator = MockModelIterator(mock_example)
        iterator.settings['max_invalid_attempts'] = 3
        
        # Mock to always return invalid models
        with patch.object(iterator.model_builder, 'build_new_model_structure') as mock_build:
            mock_structure = Mock()
            mock_structure.z3_world_states = []
            mock_build.return_value = mock_structure
            
            # Run iteration
            iterator.max_iterations = 10
            models = iterator.iterate()
            
            # Should stop after max_invalid_attempts
            assert len(models) == 1
            assert any("Too many consecutive invalid" in msg 
                      for msg in iterator.debug_messages)
    
    def test_isomorphism_detection(self):
        """Test isomorphism detection and escape attempts."""
        # This test requires NetworkX
        pytest.importorskip("networkx")
        
        mock_example = create_mock_example()
        iterator = MockModelIterator(mock_example)
        
        # TODO: Implement isomorphism test
        # Requires mocking ModelGraph and isomorphism checks
    
    def test_debug_message_collection(self):
        """Test debug messages are properly collected."""
        mock_example = create_mock_example()
        iterator = MockModelIterator(mock_example)
        
        # Run single iteration
        iterator.max_iterations = 2
        models = iterator.iterate()
        
        # Check debug messages
        debug_msgs = iterator.get_debug_messages()
        # Should have collected some debug messages during iteration
        assert len(debug_msgs) > 0
        # Messages should be strings describing iteration events
        assert all(isinstance(msg, str) for msg in debug_msgs)
        # The test should check for actual messages produced during iteration
        # Since the mock returns None for model building, we should see those messages
        assert any("failed to build" in msg.lower() or "no model" in msg.lower() or "unsat" in msg.lower() for msg in debug_msgs)


def create_mock_example():
    """Create a mock BuildExample for testing."""
    mock_example = Mock(spec=BuildExample)
    
    # Mock model structure
    mock_structure = Mock()
    mock_structure.z3_model_status = True
    mock_structure.z3_model = Mock()
    mock_structure.solver = z3.Solver()
    mock_structure.all_states = [z3.BitVecVal(i, 4) for i in range(4)]
    mock_structure.z3_world_states = [z3.BitVecVal(0, 4)]
    mock_structure.z3_possible_states = [z3.BitVecVal(0, 4), z3.BitVecVal(1, 4)]
    mock_structure.sentence_letters = []
    mock_structure.semantics = Mock()
    
    # Mock model constraints (required for constraint preservation)
    mock_constraints = Mock()
    mock_constraints.all_constraints = [z3.BoolVal(True), z3.BoolVal(True)]  # Mock constraints
    
    mock_example.model_structure = mock_structure
    mock_example.model_constraints = mock_constraints
    mock_example.settings = {'iterate': 5, 'max_time': 1.0}
    
    return mock_example


"""Tests for the modal subtheory.

This package contains tests for the modal operators including necessity (□)
and possibility (◇), as well as their interaction with other operators.

The test suite includes:
- Example tests validating logical properties of modal operators
- Unit tests for operator implementations
- Tests for modal axioms (K, T, 4, 5, etc.)
- Integration tests with extensional operators

For running these tests:
    # All modal tests
    pytest src/model_checker/theory_lib/logos/subtheories/modal/tests/
    
    # Specific test file
    pytest src/model_checker/theory_lib/logos/subtheories/modal/tests/test_modal_examples.py
    
    # With verbose output
    pytest -v src/model_checker/theory_lib/logos/subtheories/modal/tests/
"""


"""
Expression parsing utilities.

This module provides functions for parsing logical expressions from infix
notation to the prefix notation used internally by the framework.
"""


def parse_expression(tokens):
    """Parses a list of tokens representing a propositional expression and returns
    the expression in prefix notation.
    """
    if not tokens:  # Check if tokens are empty before processing
        raise ValueError("Empty token list")
    token = tokens.pop(0)  # Get the next token
    if token == "(":  # Handle binary operator case (indicated by parentheses)
        if not tokens:  # Check if tokens are empty after removing first token
            raise ValueError(f"Empty token list after opening parenthesis")
        closing_parentheses = tokens.pop()  # Remove the closing parenthesis
        if closing_parentheses != ")":
            raise SyntaxError(
                f"The sentence {tokens} is missing a closing parenthesis."
            )
        if not tokens:  # Check if tokens are empty after removing parentheses
            raise ValueError(f"Empty token list after removing parentheses")
        operator, left, right = op_left_right(tokens)
        left_arg, left_comp = parse_expression(left)  # Parse the left argument
        right_arg, right_comp = parse_expression(right)  # Parse the right argument
        complexity = left_comp + right_comp + 1
        return [operator, left_arg, right_arg], complexity 
    if token.isalnum():  # Handle atomic sentences
        return [token], 0
    elif token.startswith("\\"):  # Handle all LaTeX operators 
        # This includes: \top, \bot, \Future, \Past, \future, \past, etc.
        # Special case for nullary operators like \top and \bot
        if token in {"\\top", "\\bot"}:
            return [token], 0
        if not tokens:  # Check if tokens are empty after operator
            raise ValueError(f"Empty token list after operator {token}")
        arg, comp = parse_expression(tokens)
        return [token, arg], comp + 1
    arg, comp = parse_expression(tokens)
    return [token, arg], comp + 1 


def op_left_right(tokens):
    """Divides whatever is inside a pair of parentheses into the left argument,
    right argument, and the operator."""

    def balanced_parentheses(tokens):
        """Check if parentheses are balanced in the argument."""
        stack = []
        for token in tokens:
            if token == "(":
                stack.append(token)
            elif token == ")":
                if not stack:
                    return False
                stack.pop()
        return len(stack) == 0

    def check_right(tokens, operator):
        if not tokens:
            raise ValueError(f"Expected an argument after the operator {operator}")
        if not balanced_parentheses(tokens):
            raise ValueError(
                f"Unbalanced parentheses for the tokens {tokens} " + 
                f"with the operator {operator}."
            )
        return tokens  # Remaining tokens are the right argument

    def cut_parentheses(left, tokens):
        count = 1  # To track nested parentheses
        while tokens:
            token = tokens.pop(0)
            if token == "(":
                count += 1
                left.append(token)
            elif token == ")":
                count -= 1
                left.append(token)
            elif count > 0:
                left.append(token)
            elif not tokens:
                raise ValueError(f"Extra parentheses in {tokens}.")
            else:
                operator = token
                right = check_right(tokens, operator)
                return operator, left, right
        raise ValueError(f"The expression {tokens} is incomplete.")

    def process_operator(tokens):
        if tokens:
            return tokens.pop(0)
        raise ValueError("Operator missing after operand")

    def extract_arguments(tokens):
        """Extracts the left argument and right argument from tokens."""
        left = []
        while tokens:
            token = tokens.pop(0)
            if token == "(":
                left.append(token)
                return cut_parentheses(left, tokens)
            elif token.isalnum() or token in {"\\top", "\\bot"}:
                left.append(token)
                operator = process_operator(tokens)
                right = check_right(tokens, operator)
                return operator, left, right
            else:
                left.append(token)
        raise ValueError("Invalid expression or unmatched parentheses")

    result = extract_arguments(tokens)
    if result is None:
        raise ValueError("Failed to extract arguments")
    return result


"""
Metadata management for theory libraries.

This module provides utilities for managing version information and licensing
across all theory libraries in the ModelChecker framework.

It includes functions for:
- Updating version information
- Creating/updating license files
- Creating/updating citation files
- Verifying metadata consistency
"""

import os
import re
import datetime
from typing import Dict, List, Optional, Any, Tuple

from model_checker.utils import get_model_checker_version

from model_checker.theory_lib import (
    AVAILABLE_THEORIES,
    get_theory_version_registry,
    get_theory_license_info,
    create_license_file,
    create_citation_file
)

def update_all_theory_versions(new_version: Optional[str] = None) -> Dict[str, str]:
    """Update version information for all theories.
    
    Args:
        new_version: Optional new version to set for all theories.
                    If None, only updates the model_checker_version.
    
    Returns:
        dict: Dictionary mapping theory names to their updated versions
    """
    updated_versions = {}
    model_checker_version = get_model_checker_version()
    
    for theory_name in AVAILABLE_THEORIES:
        try:
            # Get theory directory
            theory_dir = os.path.dirname(os.path.abspath(__file__))
            init_path = os.path.join(theory_dir, theory_name, "__init__.py")
            
            if not os.path.exists(init_path):
                print(f"WARNING: __init__.py not found for theory '{theory_name}'")
                continue
                
            # Read current content
            with open(init_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Update model_checker_version
            mc_version_pattern = re.compile(
                r'__model_checker_version__\s*=\s*["\'].*?["\']'
            )
            
            # Check if the file contains model_checker_version
            if '__model_checker_version__' in content:
                content = mc_version_pattern.sub(
                    f'__model_checker_version__ = "{model_checker_version}"', 
                    content
                )
            else:
                # Add the import for get_model_checker_version if needed
                if 'from model_checker.utils import get_model_checker_version' not in content:
                    import_pattern = re.compile(r'(from[\s\S]*?\n\n|import[\s\S]*?\n\n)')
                    # Try to add after imports
                    if import_pattern.search(content):
                        content = import_pattern.sub(
                            r'\1# Import version utilities\nfrom model_checker.utils import get_model_checker_version\n\n',
                            content
                        )
                    else:
                        # Add at the top after any docstring
                        if content.startswith('"""'):
                            # Find the end of the docstring
                            end_docstring = content.find('"""', 3)
                            if end_docstring > 0:
                                insertion_point = end_docstring + 3
                                content = content[:insertion_point] + '\n\n# Import version utilities\nfrom model_checker.utils import get_model_checker_version\n' + content[insertion_point:]
                        else:
                            # Just add at the top
                            content = '# Import version utilities\nfrom model_checker.utils import get_model_checker_version\n\n' + content
                
                # Find a good place to add version information
                version_block = f"\n# Version information\n__model_checker_version__ = \"{model_checker_version}\"  # ModelChecker version this was built with\n"
                
                # Add before __all__ if it exists
                all_pattern = re.compile(r'__all__\s*=\s*\[')
                if all_pattern.search(content):
                    all_match = all_pattern.search(content)
                    if all_match:
                        match_pos = all_match.start()
                        content = content[:match_pos] + version_block + content[match_pos:]
                else:
                    # Add at the end
                    content += version_block
            
            # Update theory version if specified
            if new_version:
                version_pattern = re.compile(r'__version__\s*=\s*["\'].*?["\']')
                
                if '__version__' in content:
                    content = version_pattern.sub(f'__version__ = "{new_version}"', content)
                else:
                    # Find where to insert the version
                    if '__model_checker_version__' in content:
                        content = content.replace(
                            f'__model_checker_version__ = "{model_checker_version}"',
                            f'__version__ = "{new_version}"  # Theory version\n__model_checker_version__ = "{model_checker_version}"'
                        )
                    else:
                        # Add at the end
                        content += f'\n__version__ = "{new_version}"  # Theory version\n'
                
                updated_versions[theory_name] = new_version
            else:
                # Extract current version
                version_pattern = re.compile(r'__version__\s*=\s*["\'](.+?)["\']')
                version_match = version_pattern.search(content)
                if version_match:
                    updated_versions[theory_name] = version_match.group(1)
                else:
                    updated_versions[theory_name] = "unknown"
            
            # Update __all__ list to include version attributes if needed
            if '__all__' in content:
                all_content_pattern = re.compile(r'__all__\s*=\s*\[([\s\S]*?)\]')
                all_content_match = all_content_pattern.search(content)
                if all_content_match:
                    all_items = all_content_match.group(1)
                    updates_needed = []
                    
                    if '"__version__"' not in all_items and "'__version__'" not in all_items:
                        updates_needed.append('"__version__",                # Theory version information')
                    
                    if '"__model_checker_version__"' not in all_items and "'__model_checker_version__'" not in all_items:
                        updates_needed.append('"__model_checker_version__",  # Compatible ModelChecker version')
                    
                    if updates_needed:
                        # Add the missing items to __all__
                        updated_all = all_items.rstrip()
                        if not updated_all.endswith(','):
                            updated_all += ','
                        
                        for item in updates_needed:
                            updated_all += f'\n    {item}'
                        
                        content = all_content_pattern.sub(f"__all__ = [{updated_all}\n]", content)
            
            # Write updated content
            with open(init_path, 'w', encoding='utf-8') as f:
                f.write(content)
            
        except Exception as e:
            print(f"Error updating version for theory '{theory_name}': {str(e)}")
            updated_versions[theory_name] = "error"
    
    return updated_versions

def create_all_license_files(license_type: str = "GPL-3.0", author_info: Optional[Dict[str, Any]] = None) -> Dict[str, bool]:
    """Create license files for all theories.
    
    Args:
        license_type: Type of license (GPL-3.0, MIT, etc.)
        author_info: Optional author information
    
    Returns:
        dict: Dictionary mapping theory names to success status
    """
    results = {}
    
    for theory_name in AVAILABLE_THEORIES:
        try:
            result = create_license_file(theory_name, license_type, author_info)
            results[theory_name] = result
        except Exception as e:
            print(f"Error creating license for theory '{theory_name}': {str(e)}")
            results[theory_name] = False
    
    return results

def create_all_citation_files(author_info: Optional[Dict[str, str]] = None) -> Dict[str, bool]:
    """Create citation files for all theories.
    
    Args:
        author_info: Optional author information with keys 'name', 'email', etc.
    
    Returns:
        dict: Dictionary mapping theory names to success status
    """
    results = {}
    year = datetime.datetime.now().year
    
    # Get author name from info or use placeholder
    author_name = author_info.get('name', '[Author Name]') if author_info else '[Author Name]'
    
    for theory_name in AVAILABLE_THEORIES:
        try:
            # Generate citation text
            citation_text = f"""# Citation Information

If you use this theory implementation in academic work, please cite:

{author_name}. ({year}). {theory_name.title()}: A semantic theory implementation for the
ModelChecker framework.

## Theory Implementation Notes

This theory implementation is part of the ModelChecker framework.
For more detailed information about the theory's mathematical foundations,
please include additional references here.
"""
            result = create_citation_file(theory_name, citation_text)
            results[theory_name] = result
        except Exception as e:
            print(f"Error creating citation for theory '{theory_name}': {str(e)}")
            results[theory_name] = False
    
    return results

def verify_metadata_consistency() -> Dict[str, Dict[str, Any]]:
    """Verify metadata consistency across all theories.
    
    Checks for consistent versioning, licensing, and citation information.
    
    Returns:
        dict: Dictionary with verification results for each theory
    """
    results = {}
    
    for theory_name in AVAILABLE_THEORIES:
        theory_results = {
            'version': {
                'found': False,
                'value': None,
                'model_checker_version_found': False,
                'model_checker_version': None,
            },
            'license': {
                'found': False,
                'type': None,
            },
            'citation': {
                'found': False,
            }
        }
        
        # Check for version info
        try:
            # Import the theory module
            try:
                import importlib
                theory_module = importlib.import_module(f"model_checker.theory_lib.{theory_name}")
                
                # Check version
                if hasattr(theory_module, "__version__"):
                    theory_results['version']['found'] = True
                    theory_results['version']['value'] = theory_module.__version__
                
                # Check model_checker_version
                if hasattr(theory_module, "__model_checker_version__"):
                    theory_results['version']['model_checker_version_found'] = True
                    theory_results['version']['model_checker_version'] = theory_module.__model_checker_version__
            except ImportError:
                pass
        except Exception as e:
            theory_results['version']['error'] = str(e)
        
        # Check for license file
        try:
            license_info = get_theory_license_info(theory_name)
            theory_results['license']['found'] = license_info['path'] is not None
            theory_results['license']['type'] = license_info['type']
        except Exception as e:
            theory_results['license']['error'] = str(e)
        
        # Check for citation file
        try:
            theory_dir = os.path.dirname(os.path.abspath(__file__))
            citation_path = os.path.join(theory_dir, theory_name, "CITATION.md")
            theory_results['citation']['found'] = os.path.exists(citation_path)
        except Exception as e:
            theory_results['citation']['error'] = str(e)
        
        results[theory_name] = theory_results
    
    return results

def print_metadata_report(verify_output: Optional[Dict[str, Dict[str, Any]]] = None) -> None:
    """Print a formatted report of metadata consistency.
    
    Args:
        verify_output: Optional verification output from verify_metadata_consistency()
                      If None, will call the function to get fresh results
    """
    if verify_output is None:
        verify_output = verify_metadata_consistency()
    
    print("\n=== ModelChecker Theory Metadata Report ===\n")
    print(f"ModelChecker Version: {get_model_checker_version()}")
    print(f"Theories: {', '.join(AVAILABLE_THEORIES)}")
    print("\n--- Metadata Status ---\n")
    
    # Calculate column widths
    theory_width = max(len(t) for t in AVAILABLE_THEORIES) + 2
    version_width = 12
    mc_version_width = 20
    license_width = 10
    citation_width = 10
    
    # Print header
    header = (
        f"{'Theory':{theory_width}} | "
        f"{'Version':{version_width}} | "
        f"{'MC Version':{mc_version_width}} | "
        f"{'License':{license_width}} | "
        f"{'Citation':{citation_width}}"
    )
    print(header)
    print("-" * len(header))
    
    # Print each theory's status
    for theory_name, status in verify_output.items():
        version_status = status['version']['value'] if status['version']['found'] else "Missing"
        mc_version_status = status['version']['model_checker_version'] if status['version']['model_checker_version_found'] else "Missing"
        license_status = status['license']['type'] if status['license']['found'] else "Missing"
        citation_status = "Present" if status['citation']['found'] else "Missing"
        
        row = (
            f"{theory_name:{theory_width}} | "
            f"{version_status:{version_width}} | "
            f"{mc_version_status:{mc_version_width}} | "
            f"{license_status:{license_width}} | "
            f"{citation_status:{citation_width}}"
        )
        print(row)
    
    print("\n--- Recommendations ---\n")
    
    # Generate recommendations
    recommendations = []
    
    # Check version consistency
    versions = get_theory_version_registry()
    unique_versions = set(v for v in versions.values() if v != "unknown")
    if len(unique_versions) > 1:
        recommendations.append("• Version inconsistency detected. Consider running update_all_theory_versions() to standardize versions.")
    
    # Check for missing data
    missing_versions = [t for t, s in verify_output.items() if not s['version']['found']]
    missing_mc_versions = [t for t, s in verify_output.items() if not s['version']['model_checker_version_found']]
    missing_licenses = [t for t, s in verify_output.items() if not s['license']['found']]
    missing_citations = [t for t, s in verify_output.items() if not s['citation']['found']]
    
    if missing_versions:
        recommendations.append(f"• Missing version information for: {', '.join(missing_versions)}")
    
    if missing_mc_versions:
        recommendations.append(f"• Missing model_checker_version information for: {', '.join(missing_mc_versions)}")
    
    if missing_licenses:
        recommendations.append(f"• Missing license files for: {', '.join(missing_licenses)}")
    
    if missing_citations:
        recommendations.append(f"• Missing citation files for: {', '.join(missing_citations)}")
    
    # Print recommendations
    if recommendations:
        for rec in recommendations:
            print(rec)
    else:
        print("✓ All metadata is consistent and complete.")
    
    print("\n--- How to Update ---\n")
    print("To update all theory versions:")
    print("  from model_checker.theory_lib.meta_data import update_all_theory_versions")
    print("  update_all_theory_versions('1.0.0')")
    print("\nTo create missing license files:")
    print("  from model_checker.theory_lib.meta_data import create_all_license_files")
    print("  create_all_license_files('GPL-3.0', {'name': 'Author Name'})")
    print("\nTo create missing citation files:")
    print("  from model_checker.theory_lib.meta_data import create_all_citation_files")
    print("  create_all_citation_files({'name': 'Author Name'})")
    print("")

if __name__ == "__main__":
    print_metadata_report()


"""Test IteratorBuildExample functionality.

Following TDD approach - written BEFORE implementing IteratorBuildExample.
Tests verify factory function pattern and theory-agnostic implementation.
"""

import unittest
from unittest.mock import Mock, patch
import z3
from model_checker.builder.example import BuildExample


class TestIteratorBuildExample(unittest.TestCase):
    """Test IteratorBuildExample functionality."""
    
    def setUp(self):
        """Create mock BuildExample for testing."""
        self.mock_build = Mock(spec=BuildExample)
        self.mock_build.semantic_theory = {
            'semantics': Mock,
            'model': Mock,
            'proposition': Mock
        }
        self.mock_build.semantics = Mock
        self.mock_build.proposition = Mock
        # Create mock operator collection
        mock_operators = Mock()
        mock_operators.apply_operator = Mock(return_value=[Mock()])
        self.mock_build.operators = mock_operators
        # Mock model structure class to return proper model structure
        mock_model_structure = Mock()
        mock_model_structure.premises = []
        mock_model_structure.conclusions = []
        mock_model_structure.interpret = Mock()
        mock_model_structure.solver = Mock()
        self.mock_build.model_structure_class = Mock(return_value=mock_model_structure)
        self.mock_build.premises = []  # Empty to avoid syntax parsing
        self.mock_build.conclusions = []  # Empty to avoid syntax parsing
        self.mock_build.settings = {'N': 3}
        self.mock_build.settings_manager = Mock()
        self.mock_build.build_module = Mock()
        # Add missing model_constraints attribute
        self.mock_build.model_constraints = Mock()
        self.mock_build.model_constraints.semantics = Mock()
    
    @patch('model_checker.iterate.build_example.Syntax')
    @patch('model_checker.iterate.build_example.ModelConstraints')
    def test_create_with_z3_model(self, mock_constraints_class, mock_syntax_class):
        """Test factory function creates proper instance."""
        mock_z3_model = Mock()
        mock_syntax = Mock()
        mock_syntax_class.return_value = mock_syntax
        mock_constraints = Mock()
        mock_constraints_class.return_value = mock_constraints
        
        # Create instance
        from model_checker.iterate.build_example import create_with_z3_model
        iter_build = create_with_z3_model(
            self.mock_build, mock_z3_model
        )
        
        # Verify instance created
        from model_checker.iterate.build_example import IteratorBuildExample
        self.assertIsInstance(iter_build, IteratorBuildExample)
        
        # Verify model structure created
        self.assertIsNotNone(iter_build.model_structure)
        
        # Verify Z3 model stored
        self.assertEqual(iter_build.model_structure.z3_model, mock_z3_model)
    
    def test_theory_agnostic_implementation(self):
        """Verify no theory concepts in IteratorBuildExample."""
        import inspect
        
        # This test will fail until we implement the class
        try:
            from model_checker.iterate.build_example import IteratorBuildExample
            source = inspect.getsource(IteratorBuildExample)
            
            # Theory concepts that should NOT appear
            forbidden = [
                'is_world', 'possible', 'verify', 'falsify',
                'states', '2**', 'atom', 'sentence_letter'
            ]
            
            for concept in forbidden:
                self.assertNotIn(concept, source,
                    f"Theory concept '{concept}' found in IteratorBuildExample")
        except ImportError:
            # Module doesn't exist yet - this is expected in TDD
            pass
    
    @patch('model_checker.iterate.build_example.Syntax')
    @patch('model_checker.iterate.build_example.ModelConstraints')
    def test_injection_delegation(self, mock_constraints_class, mock_syntax_class):
        """Test that injection is delegated to ModelConstraints."""
        mock_z3_model = Mock()
        mock_constraints = Mock()
        mock_constraints_class.return_value = mock_constraints
        mock_syntax = Mock()
        mock_syntax_class.return_value = mock_syntax
        
        # Create instance
        from model_checker.iterate.build_example import create_with_z3_model
        iter_build = create_with_z3_model(
            self.mock_build, mock_z3_model
        )
        
        # Verify injection was called
        mock_constraints.inject_z3_values.assert_called_once_with(mock_z3_model, self.mock_build.model_constraints.semantics)
    
    @patch('model_checker.iterate.build_example.Syntax')
    @patch('model_checker.iterate.build_example.ModelConstraints')
    def test_attributes_copied_from_original(self, mock_constraints_class, mock_syntax_class):
        """Test that necessary attributes are copied from original BuildExample."""
        mock_z3_model = Mock()
        mock_syntax = Mock()
        mock_syntax_class.return_value = mock_syntax
        mock_constraints = Mock()
        mock_constraints_class.return_value = mock_constraints
        
        # Set specific attributes on mock
        self.mock_build.build_module = Mock()
        self.mock_build.semantic_theory = {'name': 'test'}
        self.mock_build.semantics = Mock()
        self.mock_build.proposition = Mock()
        # Create mock operator collection
        mock_operators = Mock()
        mock_operators.apply_operator = Mock(return_value=[Mock()])
        self.mock_build.operators = mock_operators
        # Mock model structure class to return proper model structure
        mock_model_structure = Mock()
        mock_model_structure.premises = []
        mock_model_structure.conclusions = []
        mock_model_structure.interpret = Mock()
        mock_model_structure.solver = Mock()
        self.mock_build.model_structure_class = Mock(return_value=mock_model_structure)
        self.mock_build.premises = ['P1', 'P2']
        self.mock_build.conclusions = ['C1']
        self.mock_build.settings = {'N': 5, 'other': 'value'}
        self.mock_build.settings_manager = Mock()
        
        from model_checker.iterate.build_example import create_with_z3_model
        iter_build = create_with_z3_model(self.mock_build, mock_z3_model)
        
        # Verify attributes copied
        self.assertEqual(iter_build.build_module, self.mock_build.build_module)
        self.assertEqual(iter_build.semantic_theory, self.mock_build.semantic_theory)
        self.assertEqual(iter_build.semantics, self.mock_build.semantics)
        self.assertEqual(iter_build.proposition, self.mock_build.proposition)
        self.assertEqual(iter_build.operators, self.mock_build.operators)
        self.assertEqual(iter_build.model_structure_class, self.mock_build.model_structure_class)
        self.assertEqual(iter_build.premises, self.mock_build.premises)
        self.assertEqual(iter_build.conclusions, self.mock_build.conclusions)
        # Settings should be a copy, not the same object
        self.assertEqual(iter_build.settings, self.mock_build.settings)
        self.assertIsNot(iter_build.settings, self.mock_build.settings)
    
    def test_no_class_methods_used(self):
        """Verify no class methods (per maintenance standards)."""
        import inspect
        
        try:
            from model_checker.iterate.build_example import IteratorBuildExample
            # Check for forbidden decorators
            for name, method in inspect.getmembers(IteratorBuildExample):
                if inspect.ismethod(method) or inspect.isfunction(method):
                    source = inspect.getsource(method)
                    self.assertNotIn('@classmethod', source,
                        f"@classmethod found in {name}")
                    self.assertNotIn('@staticmethod', source,
                        f"@staticmethod found in {name}")
        except ImportError:
            # Module doesn't exist yet - expected in TDD
            pass


if __name__ == '__main__':
    unittest.main()


"""Output manager for saving model checking results."""

import os
from datetime import datetime
from typing import Optional, List, Dict, Any


class OutputManager:
    """Manages output directory creation and file organization.
    
    This class handles the creation of output directories, manages different
    output modes (batch vs sequential), and coordinates the saving of model
    checking results in structured formats.
    
    Attributes:
        save_output: Whether output saving is enabled
        mode: Output mode ('batch' or 'sequential')
        sequential_files: For sequential mode, 'single' or 'multiple' files
        output_dir: Path to the created output directory
        models_data: List of model data dictionaries for JSON export
    """
    
    def __init__(self, save_output: bool, mode: str = 'batch', 
                 sequential_files: str = 'multiple',
                 interactive_manager=None):
        """Initialize output manager.
        
        Args:
            save_output: Whether to save output
            mode: Output mode ('batch' or 'sequential')
            sequential_files: For sequential mode, 'single' or 'multiple'
            interactive_manager: Optional InteractiveSaveManager instance
        """
        self.save_output = save_output
        self.output_dir = None
        self.models_data = []
        self.interactive_manager = interactive_manager
        
        # If interactive manager provided, use its mode
        if self.interactive_manager and hasattr(self.interactive_manager, 'mode'):
            self.mode = self.interactive_manager.mode
        else:
            self.mode = mode if mode in ['batch', 'sequential'] else 'batch'
        
        # Sequential mode options
        if self.mode == 'sequential':
            self.sequential_files = sequential_files
        else:
            self.sequential_files = None
            
        # Internal storage for batch mode
        self._batch_outputs = []
        
        # Track saved models in interactive mode
        self._interactive_saves = {}
        
    def should_save(self) -> bool:
        """Check if output should be saved.
        
        Returns:
            True if save_output is enabled, False otherwise
        """
        return self.save_output
        
    def create_output_directory(self, custom_name: Optional[str] = None):
        """Create timestamped output directory.
        
        Creates a directory with timestamp or custom name. For sequential
        mode with multiple files, also creates a 'sequential' subdirectory.
        
        Args:
            custom_name: Optional custom directory name instead of timestamp
        """
        if custom_name:
            self.output_dir = custom_name
        else:
            # Create timestamp-based directory name
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            self.output_dir = f'output_{timestamp}'
            
        # Create the directory
        os.makedirs(self.output_dir, exist_ok=True)
        
        # Create sequential subdirectory if needed
        if self.mode == 'sequential' and self.sequential_files == 'multiple':
            sequential_dir = os.path.join(self.output_dir, 'sequential')
            os.makedirs(sequential_dir, exist_ok=True)
            
    def save_example(self, example_name: str, example_data: Dict[str, Any],
                    formatted_output: str):
        """Save example based on current mode.
        
        Delegates to appropriate method based on output mode.
        
        Args:
            example_name: Name of the example
            example_data: Dictionary of model data for JSON export
            formatted_output: Formatted markdown output
        """
        if self.mode == 'batch':
            self._append_to_batch(example_name, example_data, formatted_output)
        else:  # sequential
            self._save_sequential(example_name, example_data, formatted_output)
            
    def _append_to_batch(self, example_name: str, example_data: Dict[str, Any],
                        formatted_output: str):
        """Append example to batch storage.
        
        Args:
            example_name: Name of the example
            example_data: Dictionary of model data
            formatted_output: Formatted output
        """
        self._batch_outputs.append(formatted_output)
        self.models_data.append(example_data)
        
    def _save_sequential(self, example_name: str, example_data: Dict[str, Any],
                        formatted_output: str):
        """Save example in sequential mode.
        
        Args:
            example_name: Name of the example
            example_data: Dictionary of model data
            formatted_output: Formatted output
        """
        # Add to models data
        self.models_data.append(example_data)
        
        if self.sequential_files == 'multiple':
            # Save to individual file in sequential subdirectory
            file_path = os.path.join(self.output_dir, 'sequential', f'{example_name}.md')
            with open(file_path, 'w', encoding='utf-8') as f:
                f.write(formatted_output)
        else:  # single file
            # Append to single EXAMPLES.md file
            file_path = os.path.join(self.output_dir, 'EXAMPLES.md')
            mode = 'a' if os.path.exists(file_path) else 'w'
            
            with open(file_path, mode, encoding='utf-8') as f:
                if mode == 'a':
                    f.write('\n\n---\n\n')  # separator
                f.write(formatted_output)
                
    def finalize(self):
        """Finalize output and save all files.
        
        For batch mode, writes the accumulated outputs to EXAMPLES.md.
        For interactive mode, creates summary.json.
        Always saves MODELS.json with all collected model data.
        """
        if self.mode == 'batch' and self._batch_outputs:
            # Save batch outputs to EXAMPLES.md
            examples_path = os.path.join(self.output_dir, 'EXAMPLES.md')
            with open(examples_path, 'w', encoding='utf-8') as f:
                f.write('\n\n---\n\n'.join(self._batch_outputs))
        
        # Create summary for interactive mode
        if self.mode == 'interactive':
            self._create_interactive_summary()
                
        # Save models JSON
        self._save_models_json()
        
    def _save_models_json(self):
        """Save models data to JSON file.
        
        Creates a MODELS.json file with metadata and all collected model data
        in a structured format suitable for programmatic access.
        """
        import json
        from datetime import datetime
        
        # Create data structure
        data = {
            "metadata": {
                "timestamp": datetime.now().isoformat(),
                "version": "1.0"
            },
            "models": self.models_data
        }
        
        # Save to JSON file
        json_path = os.path.join(self.output_dir, 'MODELS.json')
        with open(json_path, 'w', encoding='utf-8') as f:
            json.dump(data, f, indent=4, ensure_ascii=False)
            
    def create_example_directory(self, example_name: str) -> str:
        """Create directory for an example in interactive mode.
        
        Args:
            example_name: Name of the example
            
        Returns:
            str: Path to the created directory
        """
        if not self.output_dir:
            raise ValueError("Output directory not created yet")
            
        example_dir = os.path.join(self.output_dir, example_name)
        os.makedirs(example_dir, exist_ok=True)
        return example_dir
        
    def save_model_interactive(self, example_name: str, model_data: Dict[str, Any],
                              formatted_output: str, model_num: int):
        """Save individual model in interactive mode.
        
        Args:
            example_name: Name of the example
            model_data: Dictionary of model data
            formatted_output: Formatted markdown output
            model_num: Model number (1-based)
        """
        # Create example directory if needed
        example_dir = self.create_example_directory(example_name)
        
        # Save markdown file
        md_path = os.path.join(example_dir, f'MODEL_{model_num}.md')
        with open(md_path, 'w', encoding='utf-8') as f:
            f.write(formatted_output)
            
        # Save JSON file
        json_path = os.path.join(example_dir, f'MODEL_{model_num}.json')
        import json
        with open(json_path, 'w', encoding='utf-8') as f:
            json.dump(model_data, f, indent=4, ensure_ascii=False)
            
        # Track this save
        if example_name not in self._interactive_saves:
            self._interactive_saves[example_name] = []
        self._interactive_saves[example_name].append(model_num)
        
        # Add to models data for overall JSON
        self.models_data.append(model_data)
        
        # Display save location
        print(f"Saved to: {md_path}")
        
    def get_output_path(self, example_name: str, filename: str) -> str:
        """Get full path for output file.
        
        Args:
            example_name: Name of the example (for interactive mode)
            filename: Name of the file
            
        Returns:
            str: Full path to the file
        """
        if self.mode == 'interactive':
            example_dir = os.path.join(self.output_dir, example_name)
            return os.path.join(example_dir, filename)
        else:
            return os.path.join(self.output_dir, filename)
            
    def _create_interactive_summary(self):
        """Create summary.json for interactive mode.
        
        Summarizes what was saved for each example.
        """
        import json
        from datetime import datetime
        
        summary = {
            "metadata": {
                "timestamp": datetime.now().isoformat(),
                "mode": "interactive",
                "total_examples": len(self._interactive_saves),
                "total_models": sum(len(models) for models in self._interactive_saves.values())
            },
            "examples": {}
        }
        
        # Add example summaries
        for example_name, model_nums in self._interactive_saves.items():
            summary["examples"][example_name] = {
                "model_count": len(model_nums),
                "model_numbers": model_nums,
                "directory": example_name
            }
            
        # Save summary
        summary_path = os.path.join(self.output_dir, 'summary.json')
        with open(summary_path, 'w', encoding='utf-8') as f:
            json.dump(summary, f, indent=4, ensure_ascii=False)


"""Tests for interactive save manager."""

import pytest
from unittest.mock import Mock, patch, MagicMock
import os
import tempfile
import shutil

from model_checker.output.interactive import InteractiveSaveManager
from model_checker.output.input_provider import MockInputProvider


class TestInteractiveSaveManager:
    """Test InteractiveSaveManager functionality."""
    
    def setup_method(self):
        """Create temporary directory for tests."""
        self.temp_dir = tempfile.mkdtemp()
        self.original_cwd = os.getcwd()
        os.chdir(self.temp_dir)
        
    def teardown_method(self):
        """Clean up temporary directory."""
        os.chdir(self.original_cwd)
        shutil.rmtree(self.temp_dir)
    
    def test_initialization(self):
        """Test manager initialization."""
        input_provider = MockInputProvider([])
        manager = InteractiveSaveManager(input_provider)
        assert manager.mode is None
        assert manager.current_example is None
        assert manager.model_count == {}
        assert manager.input_provider is input_provider
        
    def test_prompt_save_mode(self):
        """Test save mode prompting."""
        # Test batch mode selection (input 'a')
        input_provider = MockInputProvider('a')
        manager = InteractiveSaveManager(input_provider)
        mode = manager.prompt_save_mode()
        assert mode == 'batch'
        assert manager.mode == 'batch'
        assert input_provider.prompts_received == ["Save all examples (a) or run in sequence (s)? "]
        
        # Test interactive mode selection (input 's')
        input_provider = MockInputProvider('s')
        manager = InteractiveSaveManager(input_provider)
        mode = manager.prompt_save_mode()
        assert mode == 'interactive'
        assert manager.mode == 'interactive'
        
        # Test invalid input defaults to batch
        input_provider = MockInputProvider('invalid')
        manager = InteractiveSaveManager(input_provider)
        mode = manager.prompt_save_mode()
        assert mode == 'batch'
        assert manager.mode == 'batch'
        
    @patch('model_checker.output.interactive.prompt_yes_no')
    def test_prompt_save_model(self, mock_yes_no):
        """Test save model prompting."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        manager.mode = 'interactive'
        
        # Test yes response
        mock_yes_no.return_value = True
        result = manager.prompt_save_model("TEST_EXAMPLE")
        assert result is True
        mock_yes_no.assert_called_with(
            "\nSave model for TEST_EXAMPLE?",
            default=True
        )
        
        # Test no response
        mock_yes_no.return_value = False
        result = manager.prompt_save_model("OTHER_EXAMPLE")
        assert result is False
        
    def test_prompt_save_model_batch_mode(self):
        """Test that batch mode always returns True without prompting."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        manager.mode = 'batch'
        
        with patch('model_checker.output.interactive.prompt_yes_no') as mock_yes_no:
            result = manager.prompt_save_model("TEST_EXAMPLE")
            assert result is True
            mock_yes_no.assert_not_called()
            
    @patch('model_checker.output.interactive.prompt_yes_no')
    def test_prompt_find_more_models(self, mock_yes_no):
        """Test prompting for additional models."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        manager.mode = 'interactive'
        
        # Test yes response
        mock_yes_no.return_value = True
        result = manager.prompt_find_more_models()
        assert result is True
        mock_yes_no.assert_called_with(
            "Find additional models?",
            default=False
        )
        
        # Test no response
        mock_yes_no.return_value = False
        result = manager.prompt_find_more_models()
        assert result is False
        
    def test_prompt_find_more_batch_mode(self):
        """Test that batch mode returns False without prompting."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        manager.mode = 'batch'
        
        with patch('model_checker.output.interactive.prompt_yes_no') as mock_yes_no:
            result = manager.prompt_find_more_models()
            assert result is False
            mock_yes_no.assert_not_called()
            
    @patch('model_checker.output.interactive.prompt_yes_no')
    @patch('builtins.print')
    def test_prompt_change_directory(self, mock_print, mock_yes_no):
        """Test directory change prompting."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        test_path = "/test/output/directory"
        
        # Test yes response
        mock_yes_no.return_value = True
        result = manager.prompt_change_directory(test_path)
        assert result is True
        
        # Check prompts
        mock_yes_no.assert_called_with(
            "Change to output directory?",
            default=False
        )
        
        # Check cd command display
        mock_print.assert_any_call(f"\nAll models saved to: {test_path}")
        mock_print.assert_any_call("To change directory, run:")
        mock_print.assert_any_call(f"  cd {test_path}")
        
    def test_track_model_count(self):
        """Test model counting functionality."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        
        # First model for example
        count = manager.get_next_model_number("EXAMPLE_1")
        assert count == 1
        assert manager.model_count["EXAMPLE_1"] == 1
        
        # Second model for same example
        count = manager.get_next_model_number("EXAMPLE_1")
        assert count == 2
        assert manager.model_count["EXAMPLE_1"] == 2
        
        # First model for different example
        count = manager.get_next_model_number("EXAMPLE_2")
        assert count == 1
        assert manager.model_count["EXAMPLE_2"] == 1
        
    def test_reset_for_new_example(self):
        """Test resetting state for new example."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        manager.current_example = "OLD_EXAMPLE"
        manager.model_count["OLD_EXAMPLE"] = 3
        
        manager.reset_for_new_example("NEW_EXAMPLE")
        assert manager.current_example == "NEW_EXAMPLE"
        assert "NEW_EXAMPLE" not in manager.model_count
        
    def test_is_interactive_mode(self):
        """Test mode checking."""
        manager = InteractiveSaveManager(MockInputProvider([]))
        
        # No mode set
        assert manager.is_interactive() is False
        
        # Batch mode
        manager.mode = 'batch'
        assert manager.is_interactive() is False
        
        # Interactive mode
        manager.mode = 'interactive'
        assert manager.is_interactive() is True
        
    def test_cancelled_mode_selection(self):
        """Test handling keyboard interrupt during mode selection."""
        # Since the implementation uses direct input(), KeyboardInterrupt would be raised
        # We'll update this test to match actual behavior
        input_provider = MockInputProvider([])  # Empty responses
        manager = InteractiveSaveManager(input_provider)
        
        # Test exhausted responses defaults to batch
        try:
            mode = manager.prompt_save_mode()
        except IndexError:
            # This is expected when mock runs out of responses
            # In real usage, this would be a KeyboardInterrupt
            pass


"""
Modal operators for necessity and possibility.

This module implements modal logical operators:
- Necessity (□)
- Possibility (◇)  
- Counterfactual Necessity (\\CFBox)
- Counterfactual Possibility (\\CFDiamond)
"""

import z3

from model_checker.utils import (
    ForAll,
    Exists,
)
from model_checker import syntactic

# Import required operators for defined operators
from ..extensional.operators import NegationOperator, TopOperator
from ..counterfactual.operators import CounterfactualOperator, MightCounterfactualOperator


class NecessityOperator(syntactic.Operator):
    """Implementation of the necessity/universal modality (□).
    
    This operator represents the modal necessity 'it is necessarily the case that',
    often written as □A. The semantics involves quantifying over all possible worlds
    in the model without restriction, checking that A is true in all of them."""

    name = "\\Box"
    arity = 1

    def true_at(self, argument, eval_point):
        """Defines truth conditions for necessity at an evaluation point."""
        sem = self.semantics
        u = z3.BitVec("t_nec_u", sem.N)
        return ForAll(
            u,
            z3.Implies(
                sem.is_world(u),
                sem.true_at(argument, {"world": u}),
            ),
        )
    
    def false_at(self, argument, eval_point):
        """Defines falsity conditions for necessity at an evaluation point."""
        sem = self.semantics
        u = z3.BitVec("t_nec_u", sem.N)
        return Exists(
            u,
            z3.And(
                sem.is_world(u),
                sem.false_at(argument, {"world": u}),
            ),
        )
    
    def extended_verify(self, state, argument, eval_point):
        """Defines verification conditions for necessity in the extended semantics."""
        return z3.And(
            state == self.semantics.null_state,
            self.true_at(argument, eval_point)
        )
    
    def extended_falsify(self, state, argument, eval_point):
        """Defines falsification conditions for necessity in the extended semantics."""
        return z3.And(
            state == self.semantics.null_state,
            self.false_at(argument, eval_point)
        )

    def find_verifiers_and_falsifiers(self, argument, eval_point):
        """Finds the verifiers and falsifiers for a necessity statement."""
        evaluate = argument.proposition.model_structure.z3_model.evaluate
        if bool(evaluate(self.true_at(argument, eval_point))):
            return {self.semantics.null_state}, set()
        if bool(evaluate(self.false_at(argument, eval_point))):
            return set(), {self.semantics.null_state}
        raise ValueError(
            f"{self.name} {argument} "
            f"is neither true nor false in the world {eval_point}."
        )
    
    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Prints modal operators with evaluation over all worlds."""
        all_worlds = sentence_obj.proposition.model_structure.z3_world_states
        self.print_over_worlds(sentence_obj, eval_point, all_worlds, indent_num, use_colors)


class PossibilityOperator(syntactic.DefinedOperator):
    """Implementation of the possibility/existential modality (◇).
    
    This operator represents the modal possibility 'it is possibly the case that',
    often written as ◇A. It is defined as the negation of necessity of negation:
    ◇A ≡ ¬□¬A.
    """
    
    name = "\\Diamond"
    arity = 1
    
    def derived_definition(self, argument):
        """Defines possibility as negation of necessity of negation."""
        return [NegationOperator, [NecessityOperator, [NegationOperator, argument]]]

    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Prints modal operators with evaluation over all worlds."""
        all_worlds = sentence_obj.proposition.model_structure.z3_world_states
        self.print_over_worlds(sentence_obj, eval_point, all_worlds, indent_num, use_colors)


class CFNecessityOperator(syntactic.DefinedOperator):
    """Implementation of counterfactual necessity (\\CFBox).
    
    This operator represents counterfactual necessity, which is defined
    using the modal necessity operator in the context of counterfactual logic.
    """
    
    name = "\\CFBox"
    arity = 1
    
    def derived_definition(self, argument):
        """Defines counterfactual necessity using modal necessity."""
        return [CounterfactualOperator, TopOperator, argument]

    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Prints modal operators with evaluation over all worlds."""
        all_worlds = sentence_obj.proposition.model_structure.z3_world_states
        self.print_over_worlds(sentence_obj, eval_point, all_worlds, indent_num, use_colors)


class CFPossibilityOperator(syntactic.DefinedOperator):
    """Implementation of counterfactual possibility (\\CFDiamond).
    
    This operator represents counterfactual possibility, which is defined
    using the modal possibility operator in the context of counterfactual logic.
    """
    
    name = "\\CFDiamond"
    arity = 1
    
    def derived_definition(self, argument):
        """Defines counterfactual possibility using modal possibility."""
        return [MightCounterfactualOperator, TopOperator, argument]

    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Prints modal operators with evaluation over all worlds."""
        all_worlds = sentence_obj.proposition.model_structure.z3_world_states
        self.print_over_worlds(sentence_obj, eval_point, all_worlds, indent_num, use_colors)


def get_operators():
    """
    Get all modal operators.
    
    Returns:
        dict: Dictionary mapping operator names to operator classes
    """
    return {
        "\\Box": NecessityOperator,
        "\\Diamond": PossibilityOperator,
        "\\CFBox": CFNecessityOperator,
        "\\CFDiamond": CFPossibilityOperator,
    }



"""
Extensional Examples Module for Logos Theory

This module provides extensional-specific examples for the logos semantic framework,
including both countermodels showing invalidity and theorems showing validity.

Example Categories:
------------------
1. Extensional Logic Countermodels (EXT_CM_*):
   - Tests for invalid extensional arguments
   - Examples showing when classical logic fails

2. Extensional Logic Theorems (EXT_TH_*):
   - Tests for valid extensional arguments
   - Classical logical properties and inference rules

Usage:
------
This module can be run directly with model-checker or dev_cli.py:

```bash
model-checker path/to/this/extensional.py
# or in development:
./dev_cli.py path/to/this/extensional.py
```

To use a specific collection of examples, modify the example_range dictionary below.
"""

# Standard imports
import sys
import os

# Add parent directories to path for proper imports
current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(current_dir)
parent_parent_dir = os.path.dirname(parent_dir)
if parent_dir not in sys.path:
    sys.path.insert(0, parent_dir)
if parent_parent_dir not in sys.path:
    sys.path.insert(0, parent_parent_dir)

# Import semantic classes
from ...semantic import (
    LogosSemantics,
    LogosProposition,
    LogosModelStructure,
)

# Import operators
from ...operators import LogosOperatorRegistry

# EXT_CM_1: CONTRADICTION (A does not entail not-A)
EXT_CM_1_premises = ['A']
EXT_CM_1_conclusions = ['\\neg A']
EXT_CM_1_settings = {
    'N': 3,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': True,
}
EXT_CM_1_example = [
    EXT_CM_1_premises,
    EXT_CM_1_conclusions,
    EXT_CM_1_settings,
]

# EXT_CM_2: AFFIRMING THE CONSEQUENT (Invalid inference)
EXT_CM_2_premises = ['B', '(A \\rightarrow B)']
EXT_CM_2_conclusions = ['A']
EXT_CM_2_settings = {
    'N': 3,
    'contingent': True,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 2,
    'expectation': True,
}
EXT_CM_2_example = [
    EXT_CM_2_premises,
    EXT_CM_2_conclusions,
    EXT_CM_2_settings,
]

# EXT_TH_1: MODUS PONENS (Valid inference)
EXT_TH_1_premises = ['A', '(A \\rightarrow B)']
EXT_TH_1_conclusions = ['B']
EXT_TH_1_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_1_example = [
    EXT_TH_1_premises,
    EXT_TH_1_conclusions,
    EXT_TH_1_settings,
]

# EXT_TH_2: MODUS TOLLENS (Valid inference)
EXT_TH_2_premises = ['\\neg B', '(A \\rightarrow B)']
EXT_TH_2_conclusions = ['\\neg A']
EXT_TH_2_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_2_example = [
    EXT_TH_2_premises,
    EXT_TH_2_conclusions,
    EXT_TH_2_settings,
]

# EXT_TH_3: CONJUNCTION ELIMINATION
EXT_TH_3_premises = ['(A \\wedge B)']
EXT_TH_3_conclusions = ['A']
EXT_TH_3_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_3_example = [
    EXT_TH_3_premises,
    EXT_TH_3_conclusions,
    EXT_TH_3_settings,
]

# EXT_TH_4: DISJUNCTION INTRODUCTION
EXT_TH_4_premises = ['A']
EXT_TH_4_conclusions = ['(A \\vee B)']
EXT_TH_4_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_4_example = [
    EXT_TH_4_premises,
    EXT_TH_4_conclusions,
    EXT_TH_4_settings,
]

# EXT_TH_5: DOUBLE NEGATION ELIMINATION
EXT_TH_5_premises = ['\\neg \\neg A']
EXT_TH_5_conclusions = ['A']
EXT_TH_5_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_5_example = [
    EXT_TH_5_premises,
    EXT_TH_5_conclusions,
    EXT_TH_5_settings,
]

# EXT_TH_6: LAW OF EXCLUDED MIDDLE
EXT_TH_6_premises = []
EXT_TH_6_conclusions = ['(A \\vee \\neg A)']
EXT_TH_6_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_6_example = [
    EXT_TH_6_premises,
    EXT_TH_6_conclusions,
    EXT_TH_6_settings,
]

# EXT_TH_7: DE MORGAN'S LAW 1
EXT_TH_7_premises = ['\\neg (A \\wedge B)']
EXT_TH_7_conclusions = ['(\\neg A \\vee \\neg B)']
EXT_TH_7_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_7_example = [
    EXT_TH_7_premises,
    EXT_TH_7_conclusions,
    EXT_TH_7_settings,
]

# EXT_TH_8: DE MORGAN'S LAW 2
EXT_TH_8_premises = ['\\neg (A \\vee B)']
EXT_TH_8_conclusions = ['(\\neg A \\wedge \\neg B)']
EXT_TH_8_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_8_example = [
    EXT_TH_8_premises,
    EXT_TH_8_conclusions,
    EXT_TH_8_settings,
]

# EXT_TH_9: BICONDITIONAL FORWARD
EXT_TH_9_premises = ['(A \\leftrightarrow B)', 'A']
EXT_TH_9_conclusions = ['B']
EXT_TH_9_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_9_example = [
    EXT_TH_9_premises,
    EXT_TH_9_conclusions,
    EXT_TH_9_settings,
]

# EXT_TH_10: BICONDITIONAL BACKWARD
EXT_TH_10_premises = ['(A \\leftrightarrow B)', 'B']
EXT_TH_10_conclusions = ['A']
EXT_TH_10_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_10_example = [
    EXT_TH_10_premises,
    EXT_TH_10_conclusions,
    EXT_TH_10_settings,
]

# EXT_TH_11: TOP IS TAUTOLOGY
EXT_TH_11_premises = []
EXT_TH_11_conclusions = ['\\top']
EXT_TH_11_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_11_example = [
    EXT_TH_11_premises,
    EXT_TH_11_conclusions,
    EXT_TH_11_settings,
]

# EXT_TH_12: EX FALSO QUODLIBET
EXT_TH_12_premises = ['\\bot']
EXT_TH_12_conclusions = ['A']
EXT_TH_12_settings = {
    'N': 3,
    'contingent': False,
    'non_null': True,
    'non_empty': True,
    'disjoint': False,
    'max_time': 1,
    'iterate': 1,
    'expectation': False,
}
EXT_TH_12_example = [
    EXT_TH_12_premises,
    EXT_TH_12_conclusions,
    EXT_TH_12_settings,
]

# Organize examples by category
countermodel_examples = {
    "EXT_CM_1": EXT_CM_1_example,
    "EXT_CM_2": EXT_CM_2_example,
}

theorem_examples = {
    "EXT_TH_1": EXT_TH_1_example,
    "EXT_TH_2": EXT_TH_2_example,
    "EXT_TH_3": EXT_TH_3_example,
    "EXT_TH_4": EXT_TH_4_example,
    "EXT_TH_5": EXT_TH_5_example,
    "EXT_TH_6": EXT_TH_6_example,
    "EXT_TH_7": EXT_TH_7_example,
    "EXT_TH_8": EXT_TH_8_example,
    "EXT_TH_9": EXT_TH_9_example,
    "EXT_TH_10": EXT_TH_10_example,
    "EXT_TH_11": EXT_TH_11_example,
    "EXT_TH_12": EXT_TH_12_example,
}

# Combine for unit_tests (used by test framework)
unit_tests = {**countermodel_examples, **theorem_examples}

# Default settings
general_settings = {
    "print_constraints": False,
    "print_impossible": True,
    "print_z3": False,
    "save_output": False,
    "maximize": False,
}

# Create operator registry for extensional theory
extensional_registry = LogosOperatorRegistry()
extensional_registry.load_subtheories(['extensional'])

# Define the semantic theory
extensional_theory = {
    "semantics": LogosSemantics,
    "proposition": LogosProposition,
    "model": LogosModelStructure,
    "operators": extensional_registry.get_operators(),
}

# Specify which theories to use
semantic_theories = {
    "Brast-McKie": extensional_theory,
}

# Specify which examples to run by default when running this module directly
# Testing with first few examples first
example_range = {
    # COUNTERMODELS
    # "EXT_CM_1": EXT_CM_1_example,  # DISJUNCTION INTRODUCTION
    "EXT_CM_2": EXT_CM_2_example,  # DISJUNCTIVE SYLLOGISM
    
    # THEOREMS
    # "EXT_TH_1": EXT_TH_1_example,  # IDENTITY REFLEXIVITY
    # "EXT_TH_2": EXT_TH_2_example,  # MODUS PONENS
    # "EXT_TH_3": EXT_TH_3_example,  # CONTRAPOSITION
    # "EXT_TH_4": EXT_TH_4_example,  # HYPOTHETICAL SYLLOGISM
    # "EXT_TH_5": EXT_TH_5_example,  # ARGUMENT FROM A AND B TO A
    # "EXT_TH_6": EXT_TH_6_example,  # CONJUNCTION SIMPLIFICATION
    # "EXT_TH_7": EXT_TH_7_example,  # CONJUNCTION INTRODUCTION
    # "EXT_TH_8": EXT_TH_8_example,  # BICONDITIONAL ELIMINATION
    # "EXT_TH_9": EXT_TH_9_example,  # DISJUNCTION COMMUTATIVITY
}

def get_examples():
    """
    Get all extensional examples.
    
    Returns:
        dict: Dictionary containing all extensional examples
    """
    return {
        'countermodels': extensional_cm_examples,
        'theorems': extensional_th_examples,
        'all': unit_tests
    }

# Make this module runnable from the command line
if __name__ == '__main__':
    import subprocess
    file_name = os.path.basename(__file__)
    subprocess.run(["model-checker", file_name], check=True, cwd=parent_parent_dir)



"""
Operators that work with witness predicates.

This module implements logical operators that can query witness predicates
from the model to determine verifiers. The key innovation is that exclusion
verification can be computed by checking the three conditions using the
witness predicates stored in the model.
"""

import z3
from model_checker.syntactic import Operator, OperatorCollection
from model_checker.utils import ForAll, Exists
from typing import List, Set, Optional
from .semantic import WitnessAwareModel


class UniNegationOperator(Operator):
    """
    UniNegation operator that queries witness predicates from the model.
    """
    
    name = "\\neg"
    arity = 1
    
    def true_at(self, arg, eval_point):
        """UniNegation is true when there's a verifier in the evaluation world."""
        x = z3.BitVec(f"ver_{self.semantics.counter}", self.semantics.N)
        self.semantics.counter += 1
        
        return Exists(
            [x],
            z3.And(
                self.extended_verify(x, arg, eval_point),
                self.semantics.is_part_of(x, eval_point["world"])
            )
        )
        
    def compute_verifiers(self, argument, model, eval_point):
        """
        Compute verifiers by querying witness predicates.
        """
        # Require witness-aware model
        assert isinstance(model, WitnessAwareModel), \
            "UniNegationOperator requires WitnessAwareModel with witness predicates"
            
        # Get verifiers of the argument
        arg_verifiers = self.semantics.extended_compute_verifiers(
            argument, model, eval_point
        )
        
        # Get formula string for witness lookup
        formula_str = f"\\neg({self.semantics._formula_to_string(argument)})"
        
        verifiers = []
        for state in range(2**self.semantics.N):
            if self._verifies_negation_with_predicates(
                state, formula_str, arg_verifiers, model
            ):
                verifiers.append(state)
                
        return verifiers
        
    def _verifies_negation_with_predicates(self, state: int, formula_str: str,
                                          arg_verifiers: List[int],
                                          model: WitnessAwareModel) -> bool:
        """
        Check if state verifies negation using witness predicates.
        """
        # Require witness predicates for this formula
        assert model.has_witness_for(formula_str), \
            f"Missing witness predicates for formula: {formula_str}"
            
        # Verify three conditions using witness predicates
        # Condition 1 & 2: For each verifier, check h and y values
        for v in arg_verifiers:
            h_v = model.get_h_witness(formula_str, v)
            y_v = model.get_y_witness(formula_str, v)
            
            assert h_v is not None and y_v is not None, \
                f"Witness values must exist for verifier {v} in formula {formula_str}"
                
            # Check condition 1: y_v ⊑ v and h_v excludes y_v
            if not self._eval_is_part_of(y_v, v, model):
                return False
            if not self._eval_excludes(h_v, y_v, model):
                return False
                
            # Check condition 2: h_v ⊑ state
            if not self._eval_is_part_of(h_v, state, model):
                return False
                
        # Check condition 3: minimality
        if not self._check_minimality(state, formula_str, arg_verifiers, model):
            return False
            
        return True
        
    def _check_minimality(self, state: int, formula_str: str,
                         arg_verifiers: List[int],
                         model: WitnessAwareModel) -> bool:
        """
        Check minimality condition using witness predicates.
        """
        # For each proper part z of state
        for z in range(2**self.semantics.N):
            if z == state:
                continue
                
            if not self._eval_is_part_of(z, state, model):
                continue
                
            # Check if all h values fit in z
            all_h_fit_in_z = True
            for v in arg_verifiers:
                h_v = model.get_h_witness(formula_str, v)
                assert h_v is not None, f"h witness must exist for verifier {v}"
                if not self._eval_is_part_of(h_v, z, model):
                    all_h_fit_in_z = False
                    break
                        
            if all_h_fit_in_z:
                # z should not satisfy condition 1
                z_satisfies_cond1 = True
                for v in arg_verifiers:
                    h_v = model.get_h_witness(formula_str, v)
                    y_v = model.get_y_witness(formula_str, v)
                    
                    assert h_v is not None and y_v is not None, \
                        f"Witness values must exist for verifier {v}"
                        
                    if not (self._eval_is_part_of(y_v, v, model) and
                           self._eval_excludes(h_v, y_v, model)):
                        z_satisfies_cond1 = False
                        break
                        
                if z_satisfies_cond1:
                    # Minimality violated
                    return False
                    
        return True
        
    def _eval_is_part_of(self, x: int, y: int, model: WitnessAwareModel) -> bool:
        """Evaluate is_part_of relation using the model."""
        x_bv = z3.BitVecVal(x, self.semantics.N)
        y_bv = z3.BitVecVal(y, self.semantics.N)
        result = model.eval(self.semantics.is_part_of(x_bv, y_bv))
        return z3.is_true(result)
        
    def _eval_excludes(self, x: int, y: int, model: WitnessAwareModel) -> bool:
        """Evaluate excludes relation using the model."""
        x_bv = z3.BitVecVal(x, self.semantics.N)
        y_bv = z3.BitVecVal(y, self.semantics.N)
        result = model.eval(self.semantics.excludes(x_bv, y_bv))
        return z3.is_true(result)
        
    def eval_fusion(self, x: int, y: int, model) -> Optional[int]:
        """Evaluate fusion operation using the model."""
        x_bv = z3.BitVecVal(x, self.semantics.N)
        y_bv = z3.BitVecVal(y, self.semantics.N)
        result = model.eval(self.semantics.fusion(x_bv, y_bv))
        if z3.is_bv_value(result):
            return result.as_long()
        return None
        
    def extended_verify(self, state, argument, eval_point):
        """
        Implement three-condition negation semantics with witness predicates.
        
        This implements the full negation semantics directly in the operator,
        making it modular and self-contained while using witness predicates
        for the existential quantification.
        """
        # Abbreviations
        sem = self.semantics
        N = sem.N
        extended_verify = sem.extended_verify
        excludes = sem.excludes
        is_part_of = sem.is_part_of
        
        # Get formula string for witness lookup
        # Handle different argument types
        if hasattr(argument, 'sentence_letter') and argument.sentence_letter is not None:
            # Z3 expression - extract the name
            if hasattr(argument.sentence_letter, 'decl') and hasattr(argument.sentence_letter.decl(), 'name'):
                arg_str = argument.sentence_letter.decl().name()
            else:
                arg_str = str(argument.sentence_letter)
        elif hasattr(argument, 'name'):
            arg_str = argument.name
        elif hasattr(argument, 'proposition'):
            arg_str = argument.proposition
        else:
            arg_str = str(argument)
            
        formula_str = f"\\neg({arg_str})"
        
        # Ensure witness predicates are registered for this formula
        if f"{formula_str}_h" not in sem.witness_registry.predicates:
            sem.witness_registry.register_witness_predicates(formula_str)
            
        # Get witness predicates for this formula - they must exist
        h_pred = sem.witness_registry.predicates[f"{formula_str}_h"]
        y_pred = sem.witness_registry.predicates[f"{formula_str}_y"]
        
        # Use witness predicates for the three-condition semantics
        x = z3.BitVec(f"wp_x_{sem.counter}", N)
        z = z3.BitVec(f"wp_z_{sem.counter}", N)
        sem.counter += 1

        return z3.And(
            # Condition 1: For every verifier x of argument, 
            # y_pred(x) is part of x and h_pred(x) excludes y_pred(x)
            ForAll([x], z3.Implies(
                extended_verify(x, argument, eval_point), 
                z3.And(
                    is_part_of(y_pred(x), x), 
                    excludes(h_pred(x), y_pred(x))
                )
            )),
            
            # Condition 2 (Upper Bound): For every verifier x of argument, 
            # h_pred(x) is part of state
            ForAll([x], z3.Implies(
                extended_verify(x, argument, eval_point), 
                is_part_of(h_pred(x), state)
            )),
            
            # Condition 3 (Least Upper Bound): state is the smallest state 
            # satisfying the UB condition
            ForAll([z], z3.Implies(
                ForAll([x], z3.Implies(
                    extended_verify(x, argument, eval_point), 
                    is_part_of(h_pred(x), z)
                )), 
                is_part_of(state, z)
            ))
        )
        
    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Print negation."""
        self.general_print(sentence_obj, eval_point, indent_num, use_colors)


class UniConjunctionOperator(Operator):
    """Conjunction operator for witness semantics."""
    
    name = "\\wedge"
    arity = 2
    
    def true_at(self, leftarg, rightarg, eval_point):
        """Conjunction is true when both arguments are true."""
        sem = self.semantics
        return z3.And(
            sem.true_at(leftarg, eval_point),
            sem.true_at(rightarg, eval_point)
        )
        
    def compute_verifiers(self, arg1, arg2, model, eval_point):
        """Standard conjunction semantics using product of verifier sets."""
        # Get verifiers for each argument by using the semantics' verifier computation
        ver1 = self.semantics.extended_compute_verifiers(arg1, model, eval_point)
        ver2 = self.semantics.extended_compute_verifiers(arg2, model, eval_point)
        
        # Use product method to compute all fusions
        return self.semantics.product(set(ver1), set(ver2))
        
    def extended_verify(self, state, arg1, arg2, eval_point):
        """Standard conjunction constraint."""
        x1 = z3.BitVec(f'conj_x1_{self.semantics.counter}', self.semantics.N)
        x2 = z3.BitVec(f'conj_x2_{self.semantics.counter}', self.semantics.N)
        self.semantics.counter += 1
        
        # Use the same structure as the main exclusion theory
        from model_checker.utils import Exists
        return Exists(
            [x1, x2],
            z3.And(
                self.semantics.fusion(x1, x2) == state,
                self.semantics.extended_verify(x1, arg1, eval_point),
                self.semantics.extended_verify(x2, arg2, eval_point),
            )
        )
        
    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Print conjunction."""
        self.general_print(sentence_obj, eval_point, indent_num, use_colors)


class UniDisjunctionOperator(Operator):
    """Disjunction operator for witness semantics."""
    
    name = "\\vee"
    arity = 2
    
    def true_at(self, leftarg, rightarg, eval_point):
        """Disjunction is true when at least one argument is true."""
        sem = self.semantics
        return z3.Or(
            sem.true_at(leftarg, eval_point), 
            sem.true_at(rightarg, eval_point)
        )
        
    def compute_verifiers(self, arg1, arg2, model, eval_point):
        """Standard disjunction semantics."""
        ver1 = self.semantics.extended_compute_verifiers(arg1, model, eval_point)
        ver2 = self.semantics.extended_compute_verifiers(arg2, model, eval_point)
        
        # Union of verifiers
        return list(set(ver1 + ver2))
        
    def extended_verify(self, state, arg1, arg2, eval_point):
        """Standard disjunction constraint."""
        return z3.Or(
            self.semantics.extended_verify(state, arg1, eval_point),
            self.semantics.extended_verify(state, arg2, eval_point)
        )
        
    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Print disjunction."""
        self.general_print(sentence_obj, eval_point, indent_num, use_colors)


class UniIdentityOperator(Operator):
    """Identity operator for witness semantics."""
    
    name = "\\equiv"
    arity = 2
    
    def true_at(self, leftarg, rightarg, eval_point):
        """Identity holds when arguments have same verifiers."""
        N = self.semantics.N
        sem = self.semantics
        x = z3.BitVec("t_id_x", N)
        return z3.And(
            ForAll([x],
                z3.Implies(
                    sem.extended_verify(x, leftarg, eval_point),
                    sem.extended_verify(x, rightarg, eval_point)
                )
            ),
            ForAll([x],
                z3.Implies(
                    sem.extended_verify(x, rightarg, eval_point),
                    sem.extended_verify(x, leftarg, eval_point)
                )
            )
        )
        
    def compute_verifiers(self, arg1, arg2, model, eval_point):
        """Standard identity semantics - verifiers when both sides have same verifiers."""
        ver1 = self.semantics.extended_compute_verifiers(arg1, model, eval_point)
        ver2 = self.semantics.extended_compute_verifiers(arg2, model, eval_point)
        
        # Identity holds at all states when verifier sets are equal
        if set(ver1) == set(ver2):
            return list(range(2**self.semantics.N))
        else:
            return []
        
    def extended_verify(self, state, arg1, arg2, eval_point):
        """Standard identity constraint."""
        x = z3.BitVec('x_id', self.semantics.N)
        return z3.And(
            z3.ForAll([x],
                z3.Implies(
                    self.semantics.extended_verify(x, arg1, eval_point),
                    self.semantics.extended_verify(x, arg2, eval_point)
                )
            ),
            z3.ForAll([x],
                z3.Implies(
                    self.semantics.extended_verify(x, arg2, eval_point),
                    self.semantics.extended_verify(x, arg1, eval_point)
                )
            )
        )
        
    def print_method(self, sentence_obj, eval_point, indent_num, use_colors):
        """Print identity."""
        self.general_print(sentence_obj, eval_point, indent_num, use_colors)


def create_operators():
    """Create operator collection for witness negation semantics."""
    return OperatorCollection(
        UniNegationOperator,
        UniConjunctionOperator,
        UniDisjunctionOperator,
        UniIdentityOperator,
    )

# Export the operator collection
witness_operators = create_operators()



"""Tests for the ImpositionModelIterator implementation.

This module tests the iteration flow and model presentation for the imposition theory.
"""

import pytest
import z3
from unittest.mock import patch, MagicMock

from model_checker.theory_lib.imposition.iterate import ImpositionModelIterator, iterate_example
from model_checker.builder.example import BuildExample


@pytest.mark.skip(reason="Implementation needed")
def test_basic_iteration():
    """Test that the ImpositionModelIterator can find multiple models."""
    # This test will be implemented with proper mocks
    pass


@pytest.mark.skip(reason="Implementation needed")
def test_imposition_specific_differences():
    """Test that the ImpositionModelIterator correctly calculates differences."""
    # This test will check specific difference detection for imposition theory
    pass


@pytest.mark.skip(reason="Implementation needed")
def test_iterate_example_function():
    """Test that the iterate_example function works correctly."""
    # This test will check the high-level iterate_example function
    pass


"""
Relevance Subtheory for Logos Theory

This subtheory implements relevance logic operators:
- Relevance (≼)

Note: This module imports from constitutive for operator dependencies.

API:
    get_operators(): Returns dictionary of operator name -> operator class mappings
    get_examples(): Returns example formulas demonstrating the operators
    RelevanceOperator: Direct access to the relevance operator class
"""

from .operators import get_operators, RelevanceOperator
from .examples import get_examples

__all__ = [
    'get_operators',
    'get_examples',
    'RelevanceOperator'
]


"""Tests for structure.py module.

Tests for ModelDefaults core Z3 solving functionality after refactoring.
Following TDD approach - these tests are written BEFORE extracting the class.
"""

import unittest
from unittest.mock import Mock, MagicMock, patch
import time
from z3 import Solver, sat, unsat, unknown, Bool, And

from model_checker.models.structure import ModelDefaults


class TestModelDefaultsStructure(unittest.TestCase):
    """Test ModelDefaults core structure and solving functionality."""
    
    def setUp(self):
        """Set up test fixtures."""
        # Mock model_constraints with proper constraint group attributes
        self.model_constraints = Mock()
        self.model_constraints.all_constraints = [Bool('test_constraint')]
        self.model_constraints.frame_constraints = []
        self.model_constraints.model_constraints = [Bool('test_constraint')]
        self.model_constraints.premise_constraints = []
        self.model_constraints.conclusion_constraints = []
        
        self.model_constraints.semantics = Mock()
        self.model_constraints.semantics.N = 3
        self.model_constraints.semantics.main_point = {'world': 0}
        self.model_constraints.semantics.all_states = [0, 1, 2]
        
        self.model_constraints.syntax = Mock()
        self.model_constraints.syntax.start_time = time.time()
        self.model_constraints.syntax.premises = []
        self.model_constraints.syntax.conclusions = []
        self.model_constraints.syntax.sentence_letters = []
        
        self.model_constraints.proposition_class = Mock()
        self.model_constraints.settings = {
            'max_time': 10,
            'expectation': False,
            'N': 3
        }
        
        # Mock settings
        self.settings = {
            'max_time': 10,
            'expectation': False,
            'N': 3
        }
        
        # Create instance with proper mocks
        with patch('model_checker.utils.Z3ContextManager.reset_context'):
            self.model_defaults = ModelDefaults(self.model_constraints, self.settings)
    
    def test_initialization(self):
        """Test that ModelDefaults initializes correctly."""
        # Check basic attributes
        self.assertEqual(self.model_defaults.model_constraints, self.model_constraints)
        self.assertEqual(self.model_defaults.settings, self.settings)
        self.assertEqual(self.model_defaults.max_time, 10)
        self.assertEqual(self.model_defaults.expectation, False)
        self.assertEqual(self.model_defaults.semantics, self.model_constraints.semantics)
        
        # Check color definitions
        self.assertIn('default', self.model_defaults.COLORS)
        self.assertIn('world', self.model_defaults.COLORS)
        self.assertIsNotNone(self.model_defaults.RESET)
        
        # Check solver state - ModelDefaults solves immediately upon construction
        # So solved should be True, timeout should be False, result should exist
        self.assertTrue(self.model_defaults.solved)
        self.assertIsNotNone(self.model_defaults.result)
    
    def test_setup_solver(self):
        """Test solver setup functionality."""
        # Test solver creation and constraint addition
        solver = self.model_defaults._setup_solver(self.model_constraints)
        
        # Check solver was created and returned
        self.assertIsNotNone(solver)
        self.assertIsInstance(solver, Solver)
        
        # Check constraints were added (by checking solver assertions)
        assertions = solver.assertions()
        self.assertGreater(len(assertions), 0)
    
    @patch('time.time')
    def test_solve_satisfiable(self, mock_time):
        """Test solving with satisfiable constraints."""
        # Mock time progression
        mock_time.side_effect = [0.0, 1.0]  # Start and end times
        
        # Create a satisfiable constraint with proper structure
        sat_constraints = Mock()
        sat_constraints.all_constraints = [Bool('true')]
        sat_constraints.frame_constraints = []
        sat_constraints.model_constraints = [Bool('true')]
        sat_constraints.premise_constraints = []
        sat_constraints.conclusion_constraints = []
        sat_constraints.semantics = self.model_constraints.semantics
        sat_constraints.proposition_class = self.model_constraints.proposition_class
        
        # Test solving
        result = self.model_defaults.solve(sat_constraints, 10)
        
        # Check results
        self.assertFalse(result[0])  # not timeout
        self.assertIsNotNone(result[1])  # model found
        self.assertTrue(result[2])  # satisfiable
        self.assertIsInstance(result[3], float)  # runtime
    
    def test_solve_unsatisfiable(self):
        """Test solving with unsatisfiable constraints."""
        # Create unsatisfiable constraints
        unsat_constraint = And(Bool('A'), Bool('A') == False)  # A and not A
        unsat_constraints = Mock()
        unsat_constraints.all_constraints = [unsat_constraint]
        unsat_constraints.frame_constraints = []
        unsat_constraints.model_constraints = [unsat_constraint]
        unsat_constraints.premise_constraints = []
        unsat_constraints.conclusion_constraints = []
        unsat_constraints.semantics = self.model_constraints.semantics
        unsat_constraints.proposition_class = self.model_constraints.proposition_class
        
        # Test solving
        result = self.model_defaults.solve(unsat_constraints, 10)
        
        # Check results
        self.assertFalse(result[0])  # not timeout
        self.assertIsNotNone(result[1])  # unsat core returned
        self.assertFalse(result[2])  # not satisfiable
    
    def test_process_solver_results_sat(self):
        """Test processing satisfiable solver results."""
        # Create mock satisfiable result
        mock_result = (False, Mock(), True, 1.5)  # not timeout, model, satisfiable, runtime
        
        # Process results
        self.model_defaults._process_solver_results(mock_result)
        
        # Check state
        self.assertTrue(self.model_defaults.solved)
        self.assertFalse(self.model_defaults.timeout)
        self.assertTrue(self.model_defaults.satisfiable)
        self.assertIsNotNone(self.model_defaults.z3_model)
        self.assertEqual(self.model_defaults.result, mock_result)
    
    def test_process_solver_results_unsat(self):
        """Test processing unsatisfiable solver results."""
        # Create mock unsatisfiable result
        mock_unsat_core = []  # Empty list represents unsat core
        mock_result = (False, mock_unsat_core, False, 0.8)  # not timeout, unsat core, not satisfiable, runtime
        
        # Process results
        self.model_defaults._process_solver_results(mock_result)
        
        # Check state
        self.assertFalse(self.model_defaults.timeout)
        self.assertFalse(self.model_defaults.z3_model_status)
        self.assertEqual(self.model_defaults.unsat_core, mock_unsat_core)
        self.assertEqual(self.model_defaults.z3_model_runtime, 0.8)
    
    def test_process_solver_results_timeout(self):
        """Test processing timeout solver results."""
        # Create mock timeout result
        mock_result = (True, None, False, 10.0)  # timeout, no model, not satisfiable, max runtime
        
        # Clear any existing z3_model first (from initialization)
        self.model_defaults.z3_model = None
        
        # Process results
        self.model_defaults._process_solver_results(mock_result)
        
        # Check state
        self.assertTrue(self.model_defaults.timeout)
        # When timeout, z3_model should be None
        self.assertIsNone(self.model_defaults.z3_model)
        self.assertFalse(self.model_defaults.z3_model_status)
        self.assertEqual(self.model_defaults.z3_model_runtime, 10.0)
    
    def test_create_result(self):
        """Test result tuple creation."""
        start_time = time.time()
        
        # Test satisfiable result
        result = self.model_defaults._create_result(False, Mock(), True, start_time)
        self.assertEqual(len(result), 4)
        self.assertFalse(result[0])  # not timeout
        self.assertIsNotNone(result[1])  # model
        self.assertTrue(result[2])  # satisfiable
        self.assertIsInstance(result[3], float)  # runtime
        
        # Test timeout result  
        result = self.model_defaults._create_result(True, None, None, start_time)
        self.assertTrue(result[0])  # timeout
        self.assertIsNone(result[1])  # no model
        self.assertIsNone(result[2])  # no satisfiability info
    
    def test_cleanup_solver_resources(self):
        """Test solver cleanup functionality."""
        # Set up solver manually
        solver = self.model_defaults._setup_solver(self.model_constraints)
        self.model_defaults.solver = solver
        self.assertIsNotNone(self.model_defaults.solver)
        
        # Test cleanup
        self.model_defaults._cleanup_solver_resources()
        
        # Check solver is cleaned up
        self.assertIsNone(self.model_defaults.solver)
        self.assertIsNone(self.model_defaults.z3_model)
    
    def test_re_solve(self):
        """Test re-solving with existing constraints."""
        # Note: solve() cleans up solver at the end, so re_solve needs a solver
        # Test re_solve without solver - should raise AssertionError
        with self.assertRaises(AssertionError):
            self.model_defaults.re_solve()
    
    def test_check_result(self):
        """Test result checking functionality."""
        # Test before solving - check_result needs solved state
        self.model_defaults.solved = False
        result = self.model_defaults.check_result()
        self.assertIsNone(result)
        
        # Set up solved state
        self.model_defaults.solved = True
        self.model_defaults.z3_model_status = True
        self.settings["expectation"] = True
        
        # Test expectation matching
        result = self.model_defaults.check_result()
        self.assertTrue(result)
        
        # Test expectation mismatch
        self.settings["expectation"] = False
        result = self.model_defaults.check_result()
        self.assertFalse(result)
    
    def test_interpret_method(self):
        """Test sentence interpretation method."""
        # Mock sentences with arguments attribute
        sentences = [Mock(), Mock()]
        for sentence in sentences:
            sentence.arguments = []  # Empty list, not None
            sentence.update_proposition = Mock()  # The actual method name
        
        # Test without z3_model - should return early
        self.model_defaults.z3_model = None
        self.model_defaults.interpret(sentences)
        
        # Check that update_proposition was NOT called (no model)
        for sentence in sentences:
            sentence.update_proposition.assert_not_called()
        
        # Now set a valid z3_model to enable interpretation
        self.model_defaults.z3_model = Mock()
        
        # Test interpretation with model
        self.model_defaults.interpret(sentences)
        
        # Check that update_proposition was called on each sentence
        for sentence in sentences:
            sentence.update_proposition.assert_called_once_with(self.model_defaults)
    
    def test_solver_isolation(self):
        """Test that solving works correctly with different constraint sets."""
        # First solve
        result1 = self.model_defaults.solve(self.model_constraints, 10)
        self.assertIsNotNone(result1)
        
        # Second solve with different constraints
        new_constraints = Mock()
        new_constraints.all_constraints = [Bool('different_constraint')]
        new_constraints.frame_constraints = []
        new_constraints.model_constraints = [Bool('different_constraint')]
        new_constraints.premise_constraints = []
        new_constraints.conclusion_constraints = []
        new_constraints.semantics = self.model_constraints.semantics
        new_constraints.proposition_class = self.model_constraints.proposition_class
        
        result2 = self.model_defaults.solve(new_constraints, 10)
        self.assertIsNotNone(result2)
        
        # Results should be independent (not identical objects)
        self.assertIsNot(result1, result2)
    
    def test_attribute_initialization_order(self):
        """Test critical attribute initialization order for iterator compatibility."""
        # This is the critical test for iterator regression prevention
        with patch('z3.Solver'):  # Mock Z3 solver to avoid actual solving
            model_defaults = ModelDefaults(self.model_constraints, self.settings)
        
        # Check that all critical attributes are properly initialized
        self.assertIsNotNone(model_defaults.model_constraints)
        self.assertIsNotNone(model_defaults.settings)
        self.assertIsNotNone(model_defaults.semantics)
        self.assertIsNotNone(model_defaults.proposition_class)
        
        # Check solver state attributes exist and have correct values after construction solve
        self.assertTrue(model_defaults.solved)  # ModelDefaults solves during construction
        self.assertIsNotNone(model_defaults.result)  # Result should be available
        
        # Check attribute types are correct
        self.assertIsInstance(model_defaults.timeout, bool)
        self.assertIn(model_defaults.satisfiable, [True, False, None])  # Boolean or None
        
        # Check main_point is properly initialized
        self.assertIsNotNone(model_defaults.main_point)
        
        # Check COLORS dictionary is properly initialized
        self.assertIsInstance(model_defaults.COLORS, dict)
        self.assertIn('world', model_defaults.COLORS)
        self.assertIn('default', model_defaults.COLORS)
        
        # Check constraint_dict is properly initialized
        self.assertIsInstance(model_defaults.constraint_dict, dict)
        # constraint_dict has all_constraints after solving
        self.assertGreaterEqual(len(model_defaults.constraint_dict), 0)


if __name__ == '__main__':
    unittest.main()


"""Tests for Z3 helper utilities."""

import pytest
import z3

from model_checker.utils.z3_helpers import ForAll, Exists


class TestForAll:
    """Test ForAll quantification function."""
    
    def test_forall_single_variable(self):
        """Test universal quantification with single variable."""
        x = z3.BitVec('x', 2)
        # ForAll x: x == 0 or x == 1 or x == 2 or x == 3
        # This should be True since we're checking all possible values
        formula = z3.Or(x == 0, x == 1, x == 2, x == 3)
        result = ForAll(x, formula)
        
        solver = z3.Solver()
        solver.add(result)
        assert solver.check() == z3.sat
    
    def test_forall_contradiction(self):
        """Test ForAll with contradictory formula."""
        x = z3.BitVec('x', 2)
        # ForAll x: x == 0 (which is false for x=1,2,3)
        formula = x == 0
        result = ForAll(x, formula)
        
        solver = z3.Solver()
        solver.add(result)
        assert solver.check() == z3.unsat
    
    def test_forall_multiple_variables(self):
        """Test ForAll with multiple variables."""
        x = z3.BitVec('x', 2)
        y = z3.BitVec('y', 2)
        # ForAll x,y: (x == y) => (x & y == x)
        # This is a tautology: if x equals y, then x AND y equals x
        formula = z3.Implies(x == y, (x & y) == x)
        result = ForAll([x, y], formula)
        
        solver = z3.Solver()
        solver.add(result)
        # This should be satisfiable since it's a tautology
        assert solver.check() == z3.sat


class TestExists:
    """Test Exists quantification function."""
    
    def test_exists_single_variable(self):
        """Test existential quantification with single variable."""
        x = z3.BitVec('x', 2)
        # Exists x: x == 2
        formula = x == 2
        result = Exists(x, formula)
        
        solver = z3.Solver()
        solver.add(result)
        assert solver.check() == z3.sat
    
    def test_exists_no_solution(self):
        """Test Exists with no solution."""
        x = z3.BitVec('x', 2)
        # Exists x: x > 3 (impossible for 2-bit value)
        formula = z3.UGT(x, 3)
        result = Exists(x, formula)
        
        solver = z3.Solver()
        solver.add(result)
        assert solver.check() == z3.unsat
    
    def test_exists_multiple_variables(self):
        """Test Exists with multiple variables."""
        x = z3.BitVec('x', 2)
        y = z3.BitVec('y', 2)
        # Exists x,y: x + y == 3
        formula = x + y == 3
        result = Exists([x, y], formula)
        
        solver = z3.Solver()
        solver.add(result)
        assert solver.check() == z3.sat


class TestMixedQuantifiers:
    """Test mixed quantifier scenarios."""
    
    def test_forall_exists_nesting(self):
        """Test ForAll and Exists interaction."""
        x = z3.BitVec('x', 2)
        
        # ForAll x: x == 0 or x == 1 or x == 2 or x == 3
        forall_result = ForAll(x, z3.Or(x == 0, x == 1, x == 2, x == 3))
        
        # Exists x: x == 1
        exists_result = Exists(x, x == 1)
        
        # Both should be satisfiable
        solver = z3.Solver()
        solver.add(z3.And(forall_result, exists_result))
        assert solver.check() == z3.sat


if __name__ == "__main__":
    pytest.main([__file__, "-v"])